"""Module for processing ECG data."""
import warnings
from typing import Callable, Dict, List, Literal, Optional, Sequence, Tuple, Union

import neurokit2 as nk
import numpy as np
import pandas as pd
from biopsykit.signals._base import _BaseProcessor
from biopsykit.utils.array_handling import find_extrema_in_radius, remove_outlier_and_interpolate, sanitize_input_1d
from biopsykit.utils.datatype_helper import (
    ECG_RESULT_DATAFRAME_COLUMNS,
    HEART_RATE_DATAFRAME_COLUMNS,
    R_PEAK_DATAFRAME_COLUMNS,
    EcgRawDataFrame,
    EcgResultDataFrame,
    HeartRatePhaseDict,
    RPeakDataFrame,
    _EcgResultDataFrame,
    _RPeakDataFrame,
    is_ecg_raw_dataframe,
    is_ecg_result_dataframe,
    is_r_peak_dataframe,
)
from scipy.stats import iqr
from tqdm.auto import tqdm

__all__ = ["EcgProcessor"]

from biopsykit.utils.exceptions import EcgProcessingError

ERROR_HANDLING = Literal["raise", "warn", "ignore"]
"""Available behavior types when dealing with ECG processing errors."""


def _hrv_process_get_hrv_types(hrv_types: Union[str, Sequence[str]]) -> Sequence[str]:
    if hrv_types is None:
        # TODO: change default parameter to 'all'
        hrv_types = ["hrv_time", "hrv_nonlinear"]
    if hrv_types == "all":
        hrv_types = list(_hrv_methods.keys())
    if isinstance(hrv_types, str):
        hrv_types = [hrv_types]

    # check whether all supplied hrv_types are valid
    for hrv_type in hrv_types:
        if hrv_type not in _hrv_methods:
            raise ValueError(f"Invalid 'hrv_types'. Must be in {list(_hrv_methods.keys())}, but got {hrv_type}")
    return hrv_types


class EcgProcessor(_BaseProcessor):
    """Class for processing ECG data."""

    def __init__(
        self,
        data: Union[EcgRawDataFrame, Dict[str, EcgRawDataFrame]],
        sampling_rate: Optional[float] = None,
        time_intervals: Optional[Union[pd.Series, Dict[str, Sequence[str]]]] = None,
        include_start: Optional[bool] = False,
    ):
        """Initialize a new ``EcgProcessor`` instance.

        To use this class simply pass data in form of a :class:`~pandas.DataFrame` (or a dict of such).
        If the data was recorded during a study that consists of multiple phases, the ECG data can be split into single
        phases by passing time information via the ``time_intervals`` parameter.

        Each instance of ``EcgProcessor`` has the following attributes:

        * ``data``: dict with raw ECG data, split into the specified phases. If data was not split the
          dictionary only has one entry, accessible by the key ``Data``
        * ``ecg_result`` : dict with ECG processing results from ``data``. Each dataframe in the dict has the
          following columns:

          * ``ECG_Raw``: Raw ECG signal
          * ``ECG_Clean``: Cleaned (filtered) ECG signal
          * ``ECG_Quality``: Quality indicator in the range of [0,1] for ECG signal quality
          * ``ECG_R_Peaks``: 1.0 where R peak was detected in the ECG signal, 0.0 else
          * ``R_Peak_Outlier``: 1.0 when a detected R peak was classified as outlier, 0.0 else
          * ``Heart_Rate``: Computed Heart rate interpolated to the length of the raw ECG signal

        * ``heart_rate``: dict with heart rate derived from ``data``. Each dataframe in the dict has the
          following columns:

          * ``Heart_Rate``: Computed heart rate for each detected R peak

        * ``rpeaks``: dict with R peak location indices derived from ``data``. Each dataframe in the dict has the
          following columns:

          * ``R_Peak_Quality``: Quality indicator in the range of [0,1] for quality of the original ECG signal
          * ``R_Peak_Idx``: Index of detected R peak in the raw ECG signal
          * ``RR_Interval``: Interval between the current and the successive R peak in seconds
          * ``R_Peak_Outlier``: 1.0 when a detected R peak was classified as outlier, 0.0 else


        You can either pass a data dictionary 'data_dict' containing ECG data or dataframe containing ECG data.
        For the latter, you can additionally supply time information via ``time_intervals`` parameter to automatically
        split the data into single phases.


        Parameters
        ----------
        data : :class:`~biopsykit.utils.datatype_helper.EcgRawDataFrame` or dict
            dataframe (or dict of such) with ECG data
        sampling_rate : float, optional
            sampling rate of recorded data in Hz
        time_intervals : dict or :class:`~pandas.Series`, optional
            time intervals indicating how ``data`` should be split.
            Can either be a :class:`~pandas.Series` with the `start` times of the single phases
            (the phase names are then derived from the index) or a dictionary with tuples indicating
            `start` and `end` times of phases (the phase names are then derived from the dict keys).
            Default: ``None`` (data is not split further)
        include_start : bool, optional
            ``True`` to include the data from the beginning of the recording to the first time interval as the
            first phase (then named ``Start``), ``False`` otherwise. Default: ``False``


        Examples
        --------
        >>> # Example using NilsPod Dataset
        >>> from biopsykit.io.nilspod import load_dataset_nilspod
        >>> from biopsykit.signals.ecg import EcgProcessor
        >>>
        >>> # path to file
        >>> file_path = "./NilsPod_TestData.bin"
        >>> # time zone of the recording (optional)
        >>> timezone = "Europe/Berlin"
        >>>
        >>> # define time intervals of the different recording phases
        >>> time_intervals = {"Part1": ("09:00", "09:30"), "Part2": ("09:30", "09:45"), "Part3": ("09:45", "10:00")}
        >>>
        >>> # load data from binary file
        >>> data, sampling_rate = load_dataset_nilspod(file_path=file_path, datastreams=['ecg'], timezone=timezone)
        >>> ecg_processor = EcgProcessor(data=data, sampling_rate=sampling_rate, time_intervals=time_intervals)

        """
        if sampling_rate is None:
            sampling_rate = 256.0
        super().__init__(
            data=data, sampling_rate=sampling_rate, time_intervals=time_intervals, include_start=include_start
        )
        for df in self.data.values():
            # make sure all data has the correct format
            is_ecg_raw_dataframe(df)

        self.ecg_result: Dict[str, EcgResultDataFrame] = {}
        """Dictionary with ECG processing result dataframes, split into different phases.

        Each dataframe is expected to be a ``EcgResultDataFrame``.

        See Also
        --------
        :obj:`~biopsykit.utils.datatype_helper.EcgResultDataFrame`
            dataframe format

        """

        self.heart_rate: HeartRatePhaseDict = {}
        """Dictionary with time-series heart rate data, split into different phases.

        See Also
        --------
        :obj:`~biopsykit.utils.datatype_helper.HeartRatePhaseDict`
            dictionary format

        """

        self.rpeaks: Dict[str, RPeakDataFrame] = {}
        """Dictionary with R peak location indices, split into different phases.

        See Also
        --------
        :obj:`~biopsykit.utils.datatype_helper.RPeakDataFrame`
            dictionary format

        """

    @property
    def ecg(self) -> Dict[str, pd.DataFrame]:
        """Return ECG signal after filtering, split into different phases.

        Returns
        -------
        dict
            dictionary with filtered ECG signal per phase

        """
        return {k: pd.DataFrame(v["ECG_Clean"]) for k, v in self.ecg_result.items()}

    @property
    def hr_result(self) -> HeartRatePhaseDict:
        """Return heart rate result from ECG processing, split into different phases.

        Returns
        -------
        dict
            dictionary with time-series heart rate per phase

        """
        return self.heart_rate

    def ecg_process(
        self,
        outlier_correction: Optional[Union[str, Sequence[str]]] = "all",
        outlier_params: Optional[Dict[str, Union[float, Sequence[float]]]] = None,
        title: Optional[str] = None,
        method: Optional[str] = None,
        errors: Optional[ERROR_HANDLING] = "raise",
    ) -> None:
        """Process ECG signal.

        The ECG processing pipeline consists of the following steps:

        * ``Filtering``: Uses :func:`~neurokit2.ecg.ecg_clean` to clean the ECG signal and prepare it
          for R peak detection
        * ``R-peak detection``: Uses :func:`~neurokit2.ecg.ecg_peaks` to find and extract R peaks.
        * ``Outlier correction`` (optional): Uses :meth:`~biopsykit.signals.ecg.EcgProcessor.correct_outlier`
          to check detected R peaks for outlier and impute removed outlier by linear interpolation.


        Parameters
        ----------
        outlier_correction : list, ``all`` or ``None``, optional
            List containing outlier correction methods to be applied. Alternatively, pass ``all`` to apply all
            available outlier correction methods, or ``None`` to not apply any outlier correction.
            See :meth:`~biopsykit.signals.ecg.EcgProcessor.outlier_corrections` to get a list of possible
            outlier correction methods. Default: ``all``
        outlier_params : dict
            Dictionary of outlier correction parameters or ``None`` for default parameters.
            See :meth:`~biopsykit.signals.ecg.EcgProcessor.outlier_params_default` for the default parameters.
            Default: ``None``
        title : str, optional
            title of ECG processing progress bar in Jupyter Notebooks or ``None`` to leave empty. Default: ``None``
        method : {'neurokit', 'hamilton', 'pantompkins', 'elgendi', ... }, optional
            method used to clean ECG signal and perform R-peak detection as defined by the ``neurokit`` library
            (see :func:`~neurokit2.ecg.ecg_clean` and :func:`~neurokit2.ecg.ecg_peaks`) or
            ``None`` to use default method (``neurokit``).
        errors : {'raise', 'warn', 'ignore'}, optional
                how to handle errors during ECG processing:

                * "raise" (default): raise an error
                * "warn": issue a warning but still return a dictionary with empty results for this data
                * "ignore": ignore errors and continue processing


        See Also
        --------
        :meth:`~biopsykit.signals.ecg.EcgProcessor.correct_outlier`
            function to perform R peak outlier correction
        :meth:`~biopsykit.signals.ecg.EcgProcessor.outlier_corrections`
            list of all available outlier correction methods
        :meth:`~biopsykit.signals.ecg.EcgProcessor.outlier_params_default`
            dictionary with default parameters for outlier correction
        :func:`~neurokit2.ecg.ecg_clean`
            neurokit method to clean ECG signal
        :func:`~neurokit2.ecg.ecg_peaks`
            neurokit method for R-peak detection


        Examples
        --------
        >>> from biopsykit.signals.ecg import EcgProcessor
        >>> # initialize EcgProcessor instance
        >>> ecg_processor = EcgProcessor(...)

        >>> # use default outlier correction pipeline
        >>> ecg_processor.ecg_process()

        >>> # don't apply any outlier correction
        >>> ecg_processor.ecg_process(outlier_correction=None)

        >>> # use custom outlier correction pipeline: only physiological and statistical outlier with custom thresholds
        >>> methods = ["physiological", "statistical"]
        >>> params = {
        >>>    'physiological': (50, 150),
        >>>    'statistical': 2.576
        >>>}
        >>> ecg_processor.ecg_process(outlier_correction=methods, outlier_params=params)

        >>> # Print available results from ECG processing
        >>> print(ecg_processor.ecg_result)
        >>> print(ecg_processor.rpeaks)
        >>> print(ecg_processor.heart_rate)

        """
        if method is None:
            method = "neurokit"

        for name, df in tqdm(self.data.items(), desc=title):
            try:
                ecg_result, rpeaks = self._ecg_process(df, method=method, phase=name)
                ecg_result, rpeaks = self.correct_outlier(
                    ecg_signal=ecg_result,
                    rpeaks=rpeaks,
                    outlier_correction=outlier_correction,
                    outlier_params=outlier_params,
                    sampling_rate=self.sampling_rate,
                )
                heart_rate = pd.DataFrame({"Heart_Rate": 60 / rpeaks["RR_Interval"]})
                rpeaks.loc[:, "Heart_Rate"] = heart_rate
                heart_rate_interpolated = nk.signal_interpolate(
                    x_values=np.squeeze(rpeaks["R_Peak_Idx"].values),
                    y_values=np.squeeze(heart_rate["Heart_Rate"].values),
                    x_new=np.arange(0, len(ecg_result["ECG_Clean"])),
                )
                ecg_result["Heart_Rate"] = heart_rate_interpolated
            except EcgProcessingError as e:
                if errors == "raise":
                    raise EcgProcessingError from e
                if errors == "warn":
                    warnings.warn(str(e))
                ecg_result, rpeaks, heart_rate = (
                    pd.DataFrame(columns=ECG_RESULT_DATAFRAME_COLUMNS),
                    pd.DataFrame(columns=R_PEAK_DATAFRAME_COLUMNS),
                    pd.DataFrame(columns=HEART_RATE_DATAFRAME_COLUMNS),
                )

            self.ecg_result[name] = ecg_result
            self.heart_rate[name] = heart_rate
            self.rpeaks[name] = rpeaks

    def _ecg_process(
        self, data: EcgRawDataFrame, method: Optional[str] = None, phase: Optional[str] = None
    ) -> Tuple[EcgResultDataFrame, RPeakDataFrame]:
        """Private method for ECG processing.

        Parameters
        ----------
        data : pd.DataFrame
            ECG data as pandas dataframe. Needs to have one column named 'ecg'
        method : {'neurokit', 'hamilton', 'pantompkins', 'elgendi', ... }, optional
            method for cleaning the ECG signal and R peak detection as defined by 'neurokit'.
            Default: ``None`` (corresponds to ``neurokit``)
        phase : str, optional
            phase name of the ECG data to be processed. Default: ``None``

        Returns
        -------
        ecg_result : :class:`~biopsykit.utils.datatype_helper.EcgResultDataFrame`, optional
            Dataframe with processed ECG signal. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`
        rpeaks : :class:`~biopsykit.utils.datatype_helper.RPeakDataFrame`, optional
            Dataframe with detected R peaks. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`

        """
        # get numpy
        ecg_signal = data["ecg"].to_numpy()
        # clean (i.e. filter) the ECG signal using the specified method
        ecg_cleaned = nk.ecg_clean(ecg_signal, sampling_rate=int(self.sampling_rate), method=method)

        # find peaks using the specified method
        # instant_peaks: array indicating where detected R peaks are in the raw ECG signal
        # rpeak_index array containing the indices of detected R peaks
        instant_peaks, rpeak_idx = nk.ecg_peaks(ecg_cleaned, sampling_rate=int(self.sampling_rate), method=method)
        rpeak_idx = rpeak_idx["ECG_R_Peaks"]
        instant_peaks = np.squeeze(instant_peaks.values)

        if len(rpeak_idx) <= 3:
            raise EcgProcessingError(f"Too few R peaks detected for phase '{phase}'. Please check your ECG signal.")

        # compute quality indicator
        quality = nk.ecg_quality(ecg_cleaned, rpeaks=rpeak_idx, sampling_rate=int(self.sampling_rate))

        # construct new dataframe
        ecg_result = pd.DataFrame(
            {
                "ECG_Raw": ecg_signal,
                "ECG_Clean": ecg_cleaned,
                "ECG_Quality": quality,
                "ECG_R_Peaks": instant_peaks,
                "R_Peak_Outlier": np.zeros(len(data)),
            },
            index=data.index,
        )

        # copy new dataframe consisting of R peaks indices (and their respective quality indicator)
        rpeaks = ecg_result.loc[ecg_result["ECG_R_Peaks"] == 1.0, ["ECG_Quality"]]
        rpeaks = rpeaks.rename(columns={"ECG_Quality": "R_Peak_Quality"})
        rpeaks.loc[:, "R_Peak_Idx"] = rpeak_idx
        # compute RR interval
        rpeaks["RR_Interval"] = np.ediff1d(rpeaks["R_Peak_Idx"], to_end=0) / self.sampling_rate
        # ensure equal length by filling the last value with the average RR interval
        rpeaks.loc[rpeaks.index[-1], "RR_Interval"] = rpeaks["RR_Interval"].mean()

        is_ecg_result_dataframe(ecg_result)
        is_r_peak_dataframe(rpeaks)

        return _EcgResultDataFrame(ecg_result), _RPeakDataFrame(rpeaks)

    @classmethod
    def outlier_corrections(cls) -> Sequence[str]:
        """Return all possible outlier correction methods.

        Currently available outlier correction methods are:

        * ``correlation``: Computes cross-correlation coefficient between every single beat and the average of
          all detected beats. Marks beats as outlier if cross-correlation coefficient is below a certain threshold.
        * ``quality``: Uses the ``ECG_Quality`` indicator from neurokit to assess signal quality. Marks beats as
          outlier if the quality indicator is below a certain threshold.
        * ``artifact``: Artifact detection based on `Berntson et al. (1990)`.
        * ``physiological``: Physiological outlier removal. Marks beats as outlier if their heart rate is above
          or below a threshold that is very unlikely to be achieved physiologically.
        * ``statistical_rr``: Statistical outlier removal based on RR intervals. Marks beats as outlier if the RR
          intervals are within the xx% highest or lowest values. Values are removed based on the z-score;
          e.g. 1.96 => 5% (2.5% highest, 2.5% lowest values); 2.576 => 1% (0.5% highest, 0.5% lowest values)
        * ``statistical_rr_diff``: Statistical outlier removal based on successive differences of RR intervals.
          Marks beats as outlier if the difference of successive RR intervals are within the xx% highest or
          lowest heart rates. Values are removed based on the z-score;
          e.g. 1.96 => 5% (2.5% highest, 2.5% lowest values); 2.576 => 1% (0.5% highest, 0.5% lowest values).

        See Also
        --------
        :meth:`~biopsykit.signals.ecg.EcgProcessor.correct_outlier`
            function to perform R peak outlier correction
        :meth:`~biopsykit.signals.ecg.EcgProcessor.outlier_params_default`
            dictionary with default parameters for outlier correction

        Returns
        -------
        list
            keys of all possible outlier correction methods

        References
        ----------
        Berntson, G. G., Quigley, K. S., Jang, J. F., & Boysen, S. T. (1990). An Approach to Artifact Identification:
        Application to Heart Period Data. *Psychophysiology*, 27(5), 586-598.
        https://doi.org/10.1111/j.1469-8986.1990.tb01982.x

        """
        return list(_outlier_correction_methods.keys())

    @classmethod
    def outlier_params_default(cls) -> Dict[str, Union[float, Sequence[float]]]:
        """Return default parameter for all outlier correction methods.

        .. note::
            The outlier correction method ``artifact`` has no threshold, but ``0.0`` is the default parameter in order
            to provide a homogenous interface

        See Also
        --------
        :meth:`~biopsykit.signals.ecg.EcgProcessor.correct_outlier`
            function to perform R peak outlier correction
        :meth:`~biopsykit.signals.ecg.EcgProcessor.outlier_corrections`
            list with available outlier correction methods

        Returns
        -------
        dict
            default parameters for outlier correction methods

        """
        return _outlier_correction_params_default

    @classmethod
    def correct_outlier(
        cls,
        ecg_processor: Optional["EcgProcessor"] = None,
        key: Optional[str] = None,
        ecg_signal: Optional[EcgResultDataFrame] = None,
        rpeaks: Optional[RPeakDataFrame] = None,
        outlier_correction: Optional[Union[str, None, Sequence[str]]] = "all",
        outlier_params: Optional[Dict[str, Union[float, Sequence[float]]]] = None,
        imputation_type: Optional[str] = None,
        sampling_rate: Optional[float] = 256.0,
    ) -> Tuple[EcgResultDataFrame, RPeakDataFrame]:
        """Perform outlier correction on the detected R peaks.

        Different methods for outlier detection are available (see
        :meth:`~biopsykit.signals.ecg.EcgProcessor.outlier_corrections()` to get a list of possible outlier
        correction methods). All outlier methods work independently on the detected R peaks, the results will be
        combined by a logical 'or'.

        RR intervals classified as outliers will be removed and imputed either using linear interpolation
        (setting ``imputation_type`` to ``linear``) or by replacing it with the average value of the 10 preceding
        and 10 succeding RR intervals (setting ``imputation_type`` to ``moving_average``).

        To use this function, either simply pass an :class:`~biopsykit.signals.ecg.EcgProcessor` object  together with
        a ``key`` indicating which phase needs to be processed should be processed or the two dataframes ``ecg_signal``
        and ``rpeaks`` resulting from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.

        Parameters
        ----------
        ecg_processor : :class:`~biopsykit.signals.ecg.EcgProcessor`, optional
            ``EcgProcessor`` object. If this argument is supplied, the ``key`` argument needs to be supplied as well
        key : str, optional
            Dictionary key of the phase to process. Needed when ``ecg_processor`` is passed as argument
        ecg_signal : :class:`~biopsykit.utils.datatype_helper.EcgResultDataFrame`, optional
            Dataframe with processed ECG signal. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`
        rpeaks : :class:`~biopsykit.utils.datatype_helper.RPeakDataFrame`, optional
            Dataframe with detected R peaks. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`
        outlier_correction : list, optional
            List containing the outlier correction methods to be applied.
            Pass ``None`` to not apply any outlier correction, ``all`` to apply all available outlier correction
            methods. See :meth:`~biopsykit.signals.ecg.EcgProcessor.outlier_corrections` to get a list of possible
            outlier correction methods.
            Default: ``all``
        outlier_params: dict, optional
            Dict of parameters to be passed to the outlier correction methods or ``None``
            to use default parameters (see :meth:`~biopsykit.signals.ecg.EcgProcessor.outlier_params_default`
            for more information).
            Default: ``None``
        imputation_type: str, optional
            Method for outlier imputation: ``linear`` for linear interpolation between the RR intervals before and
            after R peak outlier, or ``moving_average`` for average value of the
            10 preceding and 10 succeding RR intervals.
            Default: ``None`` (corresponds to ``moving_average``)
        sampling_rate : float, optional
            Sampling rate of recorded data in Hz. Not needed if ``ecg_processor`` is supplied as parameter.
            Default: 256

        Returns
        -------
        ecg_signal : :class:`~biopsykit.utils.datatype_helper.EcgResultDataFrame`
            processed ECG signal in standardized format
        rpeaks : :class:`~biopsykit.utils.datatype_helper.RPeakDataFrame`
            extracted R peaks in standardized format


        See Also
        --------
        :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process`
            function for ECG signal processing
        :meth:`~biopsykit.signals.ecg.EcgProcessor.outlier_corrections`
            list of all available outlier correction methods
        :meth:`~biopsykit.signals.ecg.EcgProcessor.outlier_params_default`
            dictionary with default parameters for outlier correction


        Examples
        --------
        >>> from biopsykit.signals.ecg import EcgProcessor
        >>> # initialize EcgProcessor instance
        >>> ecg_processor = EcgProcessor(...)

        >>> # Option 1: Use default outlier correction pipeline
        >>> ecg_signal, rpeaks = ecg_processor.correct_outlier(ecg_processor, key="Data")
        >>> print(ecg_signal)
        >>> print(rpeaks)
        >>> # Option 2: Use custom outlier correction pipeline: only physiological and statistical
        >>> # RR interval outlier with custom thresholds
        >>> methods = ["physiological", "statistical_rr"]
        >>> params = {
        >>>    'physiological': (50, 150),
        >>>    'statistical_rr': 2.576
        >>>}
        >>> ecg_signal, rpeaks = ecg_processor.correct_outlier(
        >>>                             ecg_processor, key="Data",
        >>>                             outlier_correction=methods,
        >>>                             outlier_params=params
        >>>                         )
        >>> print(ecg_signal)
        >>> print(rpeaks)

        """
        _assert_ecg_input(ecg_processor, key, ecg_signal, rpeaks)
        if ecg_processor is not None:
            ecg_signal = ecg_processor.ecg_result[key]
            rpeaks = ecg_processor.rpeaks[key]
            sampling_rate = ecg_processor.sampling_rate

        _check_dataframe_format(ecg_signal, rpeaks)

        outlier_correction, outlier_params, outlier_funcs = _get_outlier_params(outlier_correction, outlier_params)

        imputation_types = ["linear", "moving_average"]
        imputation_type = _get_imputation_type(imputation_type, imputation_types)

        # copy dataframe to mark removed beats later
        rpeaks_copy = rpeaks.copy()
        # get the last sample because it will get lost when computing the RR interval
        last_sample = rpeaks.iloc[-1]

        # initialize bool mask to mask outlier and add outlier column to rpeaks dataframe
        bool_mask = np.full(rpeaks.shape[0], False)
        rpeaks["R_Peak_Outlier"] = 0.0

        # TODO add source of different outlier methods for plotting/statistics
        for k in outlier_funcs:
            kwargs = {"ecg_signal": ecg_signal, "sampling_rate": sampling_rate}
            bool_mask = outlier_funcs[k](rpeaks, bool_mask, outlier_params[k], **kwargs)

        # mark all removed beats as outlier in the ECG dataframe
        rpeaks[bool_mask] = None
        removed_beats = rpeaks_copy["R_Peak_Idx"][rpeaks["R_Peak_Idx"].isna()]
        # mark all outlier with 1.0 in the column R_Peak_Outlier column
        rpeaks = rpeaks.fillna({"R_Peak_Outlier": 1.0})
        if ecg_signal is not None:
            # also mark outlier in the ECG signal dataframe
            ecg_signal.loc[removed_beats.index, "R_Peak_Outlier"] = 1.0

        # replace the last beat by average
        if "R_Peak_Quality" in rpeaks.columns:
            rpeaks.loc[last_sample.name] = [
                rpeaks["R_Peak_Quality"].mean(),
                last_sample["R_Peak_Idx"],
                rpeaks["RR_Interval"].mean(),
                0.0,
            ]

        # if imputation type is moving average: replace RR intervals by moving average before interpolating
        # the other columns
        if imputation_type == "moving_average":
            rpeaks["RR_Interval"] = rpeaks["RR_Interval"].fillna(
                rpeaks["RR_Interval"].rolling(21, center=True, min_periods=0).mean()
            )

        # interpolate all columns (except RR_Interval if imputation type is moving average)
        rpeaks = rpeaks.interpolate(method="linear", limit_direction="both")
        # drop duplicate R peaks (can happen during outlier correction at edge cases)
        rpeaks = rpeaks.drop_duplicates(subset="R_Peak_Idx")

        _check_dataframe_format(ecg_signal, rpeaks)

        if len(rpeaks) <= 3:
            raise EcgProcessingError("Too few peaks detected after outlier correction. Please check your ECG signal.")

        return _EcgResultDataFrame(ecg_signal), _RPeakDataFrame(rpeaks)

    @classmethod
    def correct_rpeaks(
        cls,
        ecg_processor: Optional["EcgProcessor"] = None,
        key: Optional[str] = None,
        rpeaks: Optional[RPeakDataFrame] = None,
        sampling_rate: Optional[float] = 256.0,
    ) -> pd.DataFrame:
        """Perform R peak correction algorithms to get less noisy HRV parameters.

        R peak correction comes from ``neurokit`` and is based on an algorithm by `Lipponen et al. (2019)`.

        To use this function, either simply pass an :class:`~biopsykit.signals.ecg.EcgProcessor` object together with
        a ``key`` indicating which phase needs to be processed should be processed or the dataframe ``rpeaks``
        which is a result from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.

        .. warning ::
            This algorithm might *add* additional R peaks or *remove* certain ones, so results of this function
            might **not** match with the R peaks of :meth:`~biopsykit.signals.ecg.EcgProcessor.rpeaks`. Thus, R peaks
            resulting from this function might not be used in combination with
            :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg` since R peak indices won't match.

        .. note ::
            In `BioPsyKit` this function is **not** applied to the detected R peaks during ECG signal processing but
            **only** used right before passing R peaks to :meth:`~biopsykit.signals.ecg.EcgProcessor.hrv_process()`.


        Parameters
        ----------
        ecg_processor : :class:`~biopsykit.signals.ecg.EcgProcessor`, optional
            ``EcgProcessor`` object. If this argument is supplied, the ``key`` argument needs to be supplied as well.
        key : str, optional
            Dictionary key of the phase to process. Needed when ``ecg_processor`` is passed as argument.
        rpeaks : :class:`~biopsykit.utils.datatype_helper.RPeakDataFrame`, optional
            Dataframe with detected R peaks. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.
        sampling_rate : float, optional
            Sampling rate of recorded data in Hz. Not needed if ``ecg_processor`` is supplied as parameter.
            Default: 256


        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe containing corrected R peak indices


        References
        ----------
        Lipponen, J. A., & Tarvainen, M. P. (2019). A robust algorithm for heart rate variability time series artefact
        correction using novel beat classification. *Journal of Medical Engineering and Technology*,
        43(3), 173-181. https://doi.org/10.1080/03091902.2019.1640306


        Examples
        --------
        >>> from biopsykit.signals.ecg import EcgProcessor
        >>> # initialize EcgProcessor instance
        >>> ep = EcgProcessor(...)
        >>> # correct R peak locations
        >>> rpeaks_corrected = ep.correct_rpeaks(ecg_processor, key="Data")

        """
        _assert_rpeaks_input(ecg_processor, key, rpeaks)
        if ecg_processor:
            rpeaks = ecg_processor.rpeaks[key]
            sampling_rate = ecg_processor.sampling_rate

        is_r_peak_dataframe(rpeaks)

        # fill missing RR intervals with interpolated R Peak Locations
        rpeaks_corrected = (rpeaks["RR_Interval"].cumsum() * sampling_rate).astype(int)
        rpeaks_corrected = np.append(
            rpeaks["R_Peak_Idx"].iloc[0],
            rpeaks_corrected.iloc[:-1] + rpeaks["R_Peak_Idx"].iloc[0],
        )
        _, rpeaks_corrected = nk.signal_fixpeaks(rpeaks_corrected, int(sampling_rate), iterative=False)
        rpeaks_corrected = rpeaks_corrected.astype(int)
        rpeaks_result = pd.DataFrame(rpeaks_corrected, columns=["R_Peak_Idx"])
        return rpeaks_result

    @classmethod
    def hrv_process(
        cls,
        ecg_processor: Optional["EcgProcessor"] = None,
        key: Optional[str] = None,
        rpeaks: Optional[RPeakDataFrame] = None,
        hrv_types: Optional[Sequence[str]] = None,
        correct_rpeaks: Optional[bool] = True,
        index: Optional[str] = None,
        index_name: Optional[str] = None,
        sampling_rate: Optional[float] = 256.0,
    ) -> pd.DataFrame:
        """Compute HRV parameters on the given data.

        By default, it applies R peak correction (see :meth:`~biopsykit.signals.ecg.EcgProcessor.correct_rpeaks`)
        before computing HRV parameters.

        To use this function, either simply pass an :class:`~biopsykit.signals.ecg.EcgProcessor` object together with
        a ``key`` indicating which phase needs to be processed should be processed or the dataframe ``rpeaks``
        which is a result from  :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.

        Parameters
        ----------
        ecg_processor : :class:`~biopsykit.signals.ecg.EcgProcessor`, optional
            ``EcgProcessor`` object. If this argument is supplied, the ``key`` argument needs to be supplied as well.
        key : str, optional
            Dictionary key of the phase to process. Needed when ``ecg_processor`` is passed as argument.
        rpeaks : :class:`~biopsykit.utils.datatype_helper.RPeakDataFrame`, optional
            Dataframe with detected R peaks. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.
        hrv_types: str (or list of such), optional
            list of HRV types to be computed. Must be a subset of ["hrv_time", "hrv_nonlinear", "hrv_frequency"]
            or "all" to compute all types of HRV. Refer to :func:`neurokit2.hrv.hrv` for further information on
            the available HRV parameters. Default: ``None`` (equals to ["hrv_time", "hrv_nonlinear"])
        correct_rpeaks : bool, optional
            ``True`` to apply R peak correction (using :meth:`~biopsykit.signals.ecg.EcgProcessor.correct_rpeaks()`)
            before computing HRV parameters, ``False`` otherwise. Default: ``True``
        index: str, optional
            Index of the computed HRV parameters. Used to concatenate HRV processing results from multiple phases into
            one joint dataframe later on. Default: ``None``
        index_name : str, optional
            Index name of the output dataframe. Only used if ``index`` is also supplied. Default: ``None``
        sampling_rate : float, optional
            Sampling rate of recorded data in Hz. Not needed if ``ecg_processor`` is supplied as parameter.
            Default: 256


        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with computed HRV parameters


        Examples
        --------
        >>> from biopsykit.signals.ecg import EcgProcessor
        >>> # initialize EcgProcessor instance
        >>> ecg_processor = EcgProcessor(...)

        >>> # HRV processing using default parameters (time and nonlinear), including R peak correction
        >>> hrv_output = ecg_processor.hrv_process(ecg_processor, key="Data")

        >>> # HRV processing using using all types, and without R peak correction
        >>> hrv_output = ecg_processor.hrv_process(ecg_processor, key="Data", hrv_types='all', correct_rpeaks=False)

        """
        _assert_rpeaks_input(ecg_processor, key, rpeaks)
        if ecg_processor:
            rpeaks = ecg_processor.rpeaks[key]
            sampling_rate = ecg_processor.sampling_rate

        is_r_peak_dataframe(rpeaks)

        if correct_rpeaks:
            rpeaks = cls.correct_rpeaks(rpeaks=rpeaks, sampling_rate=sampling_rate)

        hrv_types = _hrv_process_get_hrv_types(hrv_types)
        hrv_methods = {key: _hrv_methods[key] for key in hrv_types}

        # compute all HRV parameters
        list_hrv: List[pd.DataFrame] = [
            hrv_methods[key](rpeaks["R_Peak_Idx"], sampling_rate=sampling_rate) for key in hrv_methods
        ]
        # concat dataframe list
        hrv = pd.concat(list_hrv, axis=1)

        # TODO: use 'key' if index is None?
        if index is not None:
            # set index to dataframe if supplied
            hrv.index = [index]
            hrv.index.name = index_name
        return hrv

    def hrv_batch_process(self, hrv_types: Optional[Sequence[str]] = None) -> pd.DataFrame:
        """Compute HRV parameters over all phases.

        This function computes HRV parameters over all phases using
        :meth:`~biopsykit.signals.ecg.EcgProcessor.hrv_process()`.

        Parameters
        ----------
        hrv_types: str (or list of such), optional
            list of HRV types to be computed. Must be a subset of ['hrv_time', 'hrv_nonlinear', 'hrv_frequency']
            or 'all' to compute all types of HRV. Refer to :func:`neurokit2.hrv.hrv` for further information on
            the available HRV parameters. Default: ``None`` (equals to ['hrv_time', 'hrv_nonlinear'])

        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with HRV parameters over all phases

        """
        return pd.concat([self.hrv_process(self, key=key, index=key, hrv_types=hrv_types) for key in self.phases])

    @classmethod
    def ecg_estimate_rsp(
        cls,
        ecg_processor: Optional["EcgProcessor"] = None,
        key: Optional[str] = None,
        ecg_signal: Optional[EcgResultDataFrame] = None,
        rpeaks: Optional[RPeakDataFrame] = None,
        edr_type: Optional[str] = None,
        sampling_rate: Optional[float] = 256,
    ) -> pd.DataFrame:
        """Estimate respiration signal from ECG (ECG-derived respiration, EDR).

        To use this function, either simply pass an :class:`~biopsykit.signals.ecg.EcgProcessor` object  together with
        a ``key`` indicating which phase needs to be processed should be processed or the two dataframes ``ecg_signal``
        and ``rpeaks`` resulting from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.


        Parameters
        ----------
        ecg_processor : :class:`~biopsykit.signals.ecg.EcgProcessor`, optional
            ``EcgProcessor`` object. If this argument is supplied, the ``key`` argument needs to be supplied as well.
        key : str, optional
            Dictionary key of the phase to process. Needed when ``ecg_processor`` is passed as argument.
        ecg_signal : :class:`~biopsykit.utils.datatype_helper.EcgResultDataFrame`, optional
            Dataframe with processed ECG signal. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.
        rpeaks : :class:`~biopsykit.utils.datatype_helper.RPeakDataFrame`, optional
            Dataframe with detected R peaks. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`
        edr_type : {'peak_trough_mean', 'peak_trough_diff', 'peak_peak_interval'}, optional
            Method to use for estimating EDR. Must be one of 'peak_trough_mean', 'peak_trough_diff',
            or 'peak_peak_interval'. Default: 'peak_trough_mean'
        sampling_rate : float, optional
            Sampling rate of recorded data in Hz. Not needed if ``ecg_processor`` is supplied as parameter.
            Default: 256

        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with estimated respiration signal

        Examples
        --------
        >>> from biopsykit.signals.ecg import EcgProcessor
        >>> # initialize EcgProcessor instance
        >>> ecg_processor = EcgProcessor(...)

        >>> # Extract respiration signal estimated from ECG using the 'peak_trough_diff' method
        >>> rsp_signal = ecg_processor.ecg_estimate_rsp(ecg_processor, key="Data", edr_type='peak_trough_diff')

        """
        _assert_ecg_input(ecg_processor, key, ecg_signal, rpeaks)
        if ecg_processor:
            ecg_signal = ecg_processor.ecg_result[key]
            rpeaks = ecg_processor.rpeaks[key]
            sampling_rate = ecg_processor.sampling_rate

        is_ecg_result_dataframe(ecg_signal)
        is_r_peak_dataframe(rpeaks)

        if edr_type is None:
            edr_type = "peak_trough_mean"
        if edr_type not in _edr_methods:
            raise ValueError(f"`edr_type` must be one of {list(_edr_methods.keys())}, not {edr_type}")
        edr_func = _edr_methods[edr_type]

        # ensure numpy
        peaks = np.squeeze(rpeaks["R_Peak_Idx"].values)

        # find troughs (minimum 0.1s before R peak)
        troughs = find_extrema_in_radius(ecg_signal["ECG_Clean"], peaks, radius=(int(0.1 * sampling_rate), 0))
        # R peak outlier should not be included into EDR estimation
        outlier_mask = rpeaks["R_Peak_Outlier"] == 1
        # estimate raw EDR signal
        edr_signal_raw = edr_func(ecg_signal["ECG_Clean"], peaks, troughs)
        # remove R peak outlier, impute missing data, and interpolate signal to length of raw ECG signal
        edr_signal = remove_outlier_and_interpolate(edr_signal_raw, outlier_mask, peaks, len(ecg_signal))
        # Preprocessing: 10-th order Butterworth bandpass filter (0.1-0.5 Hz)
        edr_signal = nk.signal_filter(edr_signal, sampling_rate=int(sampling_rate), lowcut=0.1, highcut=0.5, order=10)

        return pd.DataFrame(edr_signal, index=ecg_signal.index, columns=["ECG_Resp"])

    @classmethod
    def rsa_process(
        cls,
        ecg_signal: EcgResultDataFrame,
        rsp_signal: pd.DataFrame,
        sampling_rate: Optional[float] = 256,
    ) -> Dict[str, float]:
        """Compute respiratory sinus arrhythmia (RSA) based on ECG and respiration signal.

        RSA is computed both via Peak-to-Trough (P2T) Porges-Bohrer method using :func:`~neurokit2.hrv.hrv_rsa`.


        Parameters
        ----------
        ecg_signal : :class:`~biopsykit.utils.datatype_helper.EcgResultDataFrame`, optional
            Dataframe with processed ECG signal. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.
        rsp_signal : pd.DataFrame
            Dataframe with 1-D raw respiration signal. Can be a 'true' respiration signal
            (e.g. from bioimpedance or Radar) or an 'estimated' respiration signal (e.g. from ECG-derived respiration).
        sampling_rate : float, optional
            Sampling rate of recorded data in Hz.
            Default: 256

        Returns
        -------
        dict
            Dictionary containing computed RSA metrics.

        See Also
        --------
        :func:`~neurokit2.hrv.hrv_rsa`
            compute respiratory sinus arrhythmia


        Examples
        --------
        >>> from biopsykit.signals.ecg import EcgProcessor
        >>> # initialize EcgProcessor instance
        >>> ecg_processor = EcgProcessor(...)

        >>> ecg_signal = ecg_processor.ecg_result['Data']
        >>> # Extract respiration signal estimated from ECG using the 'peak_trough_diff' method
        >>> rsp_signal = ecg_processor.ecg_estimate_rsp(ecg_processor, key="Data", edr_type='peak_trough_diff')
        >>> # Compute RSA from ECG and Respiration data
        >>> rsa_output = ecg_processor.rsa_process(ecg_signal, rsp_signal)

        """
        is_ecg_result_dataframe(ecg_signal)

        # ensure numpy
        rsp_signal = sanitize_input_1d(rsp_signal)
        # Process raw respiration input
        rsp_output = nk.rsp_process(rsp_signal, int(sampling_rate))[0]
        rsp_output.index = ecg_signal.index
        # Compute RSA
        return nk.hrv_rsa(ecg_signal, rsp_output, sampling_rate=int(sampling_rate))

    # @classmethod
    # def rsp_rsa_process(
    #     cls,
    #     ecg_processor: Optional["EcgProcessor"] = None,
    #     key: Optional[str] = None,
    #     ecg_signal: Optional[pd.DataFrame] = None,
    #     rpeaks: Optional[pd.DataFrame] = None,
    #     return_mean: Optional[bool] = True,
    #     index: Optional[str] = None,
    #     index_name: Optional[str] = None,
    #     sampling_rate: Optional[int] = 256,
    # ) -> pd.DataFrame:
    #     """
    #     Computes respiration rate (RSP) and RSA metrics for the given ECG signal. Per default it estimates the
    #     respiration signal using all three available EDR methods and averages the results. Optionally, estimation
    #     results from the individual methods can be returned.s
    #
    #     To use this function, either simply pass an `EcgProcessor` object together with a `key` indicating
    #     which sub-phase should be processed or the two dataframes `ecg_signal` and `rpeaks` resulting from
    #     `EcgProcessor.ecg_process()`.
    #
    #     Parameters
    #     ----------
    #     ecg_processor : EcgProcessor, optional
    #         `EcgProcessor` object. If this argument is passed, the `key` argument needs to be supplied as well
    #     key : str, optional
    #         Dictionary key of the sub-phase to process. Needed when `ecg_processor` is passed as argument
    #     ecg_signal : pd.DataFrame, optional
    #         dataframe with ECG signal. Output of `EcgProcessor.ecg_process()`
    #     rpeaks : pd.DataFrame, optional
    #         dataframe with R peaks. Output of `EcgProcessor.ecg_process()`
    #     return_mean: bool, optional
    #         Flag indicating whether to return the averaged estimation results or individual results for each EDR
    #         estimation method.
    #         Default: ``True``
    #     index: str, optional
    #         Index value of the computed RSP and RSA features. Used to concatenate dataframes from multiple sub-phases
    #         later.
    #         Default: None
    #     index_name : str, optional
    #         Index name of the output dataframe. Only used if 'index' is also supplied
    #     sampling_rate : float
    #         Sampling rate of recorded data in Hz. Not needed if ``ecg_processor`` is supplied as parameter.
    #         Default: 256
    #
    #     Returns
    #     -------
    #     pd.DataFrame
    #         dataframe of respiration rate and RSA estimation results
    #
    #     Examples
    #     --------
    #     >>> import biopsykit as ep
    #     >>> # initialize EcgProcessor instance
    #     >>> ecg_processor = ep.EcgProcessor(...)
    #
    #     >>> # Compute respiration rate and RSA. Extract respiration signal using all available
    #     >>> # methods and average the ensemble ('return_mean' is True by default)
    #     >>> rsp_signal = ecg_processor.rsp_rsa_process(ecg_processor, key="Data")
    #
    #     >>> # Compute respiration rate and RSA. Extract respiration signal using all available
    #     >>> # methods and return the single results
    #     >>> rsp_signal = ecg_processor.rsp_rsa_process(ecg_processor, key="Data", return_mean=False)
    #     """
    #
    #     _assert_ecg_input(ecg_processor, key, ecg_signal, rpeaks)
    #     if ecg_processor:
    #         ecg_signal = ecg_processor.ecg_result[key]
    #         rpeaks = ecg_processor.rpeaks[key]
    #         sampling_rate = ecg_processor.sampling_rate
    #
    #     # initialize dicts to store results
    #     rsp_rate = dict.fromkeys(_edr_methods.keys())
    #     rsa = dict.fromkeys(_edr_methods.keys())
    #
    #     for method in _edr_methods.keys():
    #         # estimate respiration signal, compute respiration signal and RSA using all three methods
    #         rsp_signal = cls.ecg_estimate_rsp(
    #             ecg_signal=ecg_signal,
    #             rpeaks=rpeaks,
    #             sampling_rate=sampling_rate,
    #             edr_type=method,
    #         )
    #         rsp_rate[method] = cls.rsp_compute_rate(rsp_signal, sampling_rate)
    #         rsa[method] = cls.rsa_process(ecg_signal, rsp_signal, sampling_rate)
    #
    #     if return_mean:
    #         # compute average respiration rate and RSA
    #         mean_resp_rate = np.mean(list(rsp_rate.values()))
    #         rsa = list(rsa.values())
    #         mean_rsa = {k: np.mean([t[k] for t in rsa]) for k in rsa[0]}
    #         mean_rsa["RSP_Rate"] = mean_resp_rate
    #         # dataframe reshaping
    #         if not index:
    #             index = "0"
    #             index_name = "Index"
    #         df_rsa = pd.DataFrame(mean_rsa, index=[index])
    #         df_rsa.index.name = index_name
    #         return df_rsa
    #     else:
    #         # dataframe reshaping
    #         df_rsa = pd.DataFrame(rsa).T
    #         df_rsa["RSP_Rate"] = rsp_rate.values()
    #         df_rsa.index.name = "Method"
    #         if index:
    #             return pd.concat([df_rsa], keys=[index], names=[index_name])
    #         return df_rsa


def _edr_peak_trough_mean(ecg: pd.Series, peaks: np.array, troughs: np.array) -> np.array:
    """Estimate respiration signal from ECG based on `peak-trough-mean` method.

    The `peak-trough-mean` method is based on computing the mean amplitude between R peaks (`peaks`) and
    minima before R peaks (`troughs`).

    Parameters
    ----------
    ecg : :class:`~pandas.Series`
        pandas series with ecg signal
    peaks : :class:`~numpy.array`
        array with peak indices
    troughs : :class:`~numpy.array`
        array with trough indices

    Returns
    -------
    :class:`~numpy.array`
        estimated raw respiration signal

    """
    peak_vals = np.array(ecg.iloc[peaks])
    trough_vals = np.array(ecg.iloc[troughs])
    return np.mean([peak_vals, trough_vals], axis=0)


def _edr_peak_trough_diff(ecg: pd.Series, peaks: np.array, troughs: np.array) -> np.array:
    """Estimate respiration signal from ECG based on `peak-trough-diff` method.

    The `peak-trough-diff` method is based on computing the amplitude difference between R peaks (`peaks`) and
    minima before R peaks (`troughs`).

    Parameters
    ----------
    ecg : :class:`~pandas.Series`
        pandas series with ecg signal
    peaks : :class:`~numpy.array`
        array with peak indices
    troughs : :class:`~numpy.array`
        array with trough indices

    Returns
    -------
    :class:`~numpy.array`
        estimated raw respiration signal

    """
    peak_vals = np.array(ecg.iloc[peaks])
    trough_vals = np.array(ecg.iloc[troughs])
    return peak_vals - trough_vals


def _edr_peak_peak_interval(ecg: pd.DataFrame, peaks: np.array, troughs: np.array) -> np.array:  # noqa: ARG001
    """Estimate respiration signal from ECG based on `peak-peak-interval` method.

    The `peak-peak-interval` method is based on computing RR intervals.

    .. note::
        To ensure the same length for the resulting array after computing successive differences
        the first value will be replaced by the mean of all RR intervals in the array

    Parameters
    ----------
    ecg : :class:`~pandas.Series`
        pandas series with ecg signal (unused but needed for consistent method signature)
    peaks : :class:`~numpy.array`
        array with peak indices
    troughs : :class:`~numpy.array`
        array with trough indices (unused but needed for consistent method signature)

    Returns
    -------
    :class:`~numpy.array`
        estimated raw respiration signal

    """
    peak_interval = np.ediff1d(peaks, to_begin=0)
    peak_interval[0] = peak_interval.mean()
    return peak_interval


def _correct_outlier_correlation(rpeaks: pd.DataFrame, bool_mask: np.array, corr_thres: float, **kwargs) -> np.array:
    """Apply outlier correction method 'correlation'.

    This function compute the cross-correlation coefficient between every single beat and the average of all detected
    beats. It marks beats as outlier if the cross-correlation coefficient is below a certain threshold.

    Parameters
    ----------
    rpeaks : :class:`~pandas.DataFrame`
        dataframe with detected R peaks. Output from :meth:`biopsykit.signals.ecg.EcgProcessor.ecg_process()`
    bool_mask : :class:`numpy.array`
        boolean array with beats marked as outlier.
        Results of this outlier correction method will be combined with the array using a logical 'or'
    corr_thres : float
        threshold for cross-correlation coefficient. Beats below that threshold will be marked as outlier
    **kwargs : additional parameters required for this outlier function, such as:

        * ecg_signal :class:`~pandas.DataFrame`
          dataframe with processed ECG signal. Output from :meth:`biopsykit.signals.ecg.EcgProcessor.ecg_process()`
        * sampling_rate : float
          sampling rate of recorded data in Hz

    Returns
    -------
    :class:`numpy.array`
        boolean array with beats marked as outlier. Logical 'or' combination of ``bool_mask`` and results from
        this algorithm

    """
    ecg_signal = kwargs.get("ecg_signal", None)
    sampling_rate = kwargs.get("sampling_rate", None)
    if any(v is None for v in [ecg_signal, sampling_rate]):
        raise ValueError(
            "Cannot apply outlier correction method 'correlation' because not all additionally required arguments "
            "were provided! Make sure you pass the following arguments: 'ecg_signal', 'sampling_rate'."
        )
    # signal outlier
    # segment individual heart beats
    heartbeats = nk.ecg_segment(ecg_signal["ECG_Clean"], rpeaks["R_Peak_Idx"], int(sampling_rate))
    heartbeats = nk.epochs_to_df(heartbeats)
    heartbeats_pivoted = heartbeats.pivot(index="Time", columns="Label", values="Signal")
    heartbeats = heartbeats.set_index("Index")
    heartbeats = heartbeats.loc[heartbeats.index.intersection(rpeaks["R_Peak_Idx"])].sort_values(by="Label")
    heartbeats = heartbeats[~heartbeats.index.duplicated()]
    heartbeats_pivoted.columns = heartbeats.index

    # compute the average over all heart beats and compute the correlation coefficient between all beats and
    # the average
    mean_beat = heartbeats_pivoted.mean(axis=1)
    heartbeats_pivoted["mean"] = mean_beat
    corr_coeff = heartbeats_pivoted.corr()["mean"].abs().sort_values(ascending=True)
    corr_coeff = corr_coeff.drop("mean")
    # compute RR intervals (in seconds) from R Peak Locations
    rpeaks["RR_Interval"] = np.ediff1d(rpeaks["R_Peak_Idx"], to_end=0) / sampling_rate

    # signal outlier: drop all beats that are below a correlation coefficient threshold
    return np.logical_or(bool_mask, rpeaks["R_Peak_Idx"].isin(corr_coeff[corr_coeff < corr_thres].index))


def _correct_outlier_quality(
    rpeaks: pd.DataFrame, bool_mask: np.array, quality_thres: float, **kwargs  # noqa: ARG001
) -> np.array:
    """Apply outlier correction method 'quality'.

    This function uses the ``ECG_Quality`` indicator from ``neurokit`` to assess signal quality. It marks beats as
    outlier if the quality indicator is below a certain threshold.


    Parameters
    ----------
    rpeaks : :class:`~pandas.DataFrame`
        dataframe with detected R peaks. Output from :meth:`biopsykit.signals.ecg.EcgProcessor.ecg_process()`
    bool_mask : :class:`numpy.array`
        boolean array with beats marked as outlier.
        Results of this outlier correction method will be combined with the array using a logical 'or'
    quality_thres : float
        threshold for signal quality indicator. Beats below that threshold will be marked as outlier


    Returns
    -------
    :class:`numpy.array`
        boolean array with beats marked as outlier. Logical 'or' combination of ``bool_mask`` and results from
        this algorithm

    """
    # signal outlier: drop all beats that are below a signal quality threshold
    return np.logical_or(bool_mask, rpeaks["R_Peak_Quality"] < quality_thres)


def _correct_outlier_statistical_rr(
    rpeaks: pd.DataFrame,
    bool_mask: np.array,
    stat_thres: float,
    **kwargs,  # noqa: ARG001
) -> np.array:
    """Apply outlier correction method 'statistical_rr'.

    This function marks beats as outlier if they are within the xx % highest or lowest RR intervals, i.e., if
    their z-score is above a threshold, e.g., ``1.96`` => 5% (2.5% highest, 2.5% lowest values);
    ``2.576`` => 1% (0.5% highest, 0.5% lowest values)


    Parameters
    ----------
    rpeaks : :class:`~pandas.DataFrame`
        dataframe with detected R peaks. Output from :meth:`biopsykit.signals.ecg.EcgProcessor.ecg_process()`
    bool_mask : :class:`numpy.array`
        boolean array with beats marked as outlier.
        Results of this outlier correction method will be combined with the array using a logical 'or'
    stat_thres : float
        threshold for z-score. Beats above that threshold will be marked as outlier


    Returns
    -------
    :class:`numpy.array`
        boolean array with beats marked as outlier. Logical 'or' combination of ``bool_mask`` and results from
        this algorithm

    """
    # statistical outlier: remove the x% highest and lowest RR intervals
    # (1.96 std = 5% outlier, 2.576 std = 1% outlier)
    rri = rpeaks["RR_Interval"]
    z_score = (rri - np.nanmean(rri)) / np.nanstd(rri, ddof=1)

    return np.logical_or(bool_mask, np.abs(z_score) > stat_thres)


def _correct_outlier_statistical_rr_diff(
    rpeaks: pd.DataFrame, bool_mask: np.array, stat_thres: float, **kwargs  # noqa: ARG001
) -> np.array:
    """Apply outlier correction method 'statistical_rr_diff'.

    This function marks beats as outlier if their successive differences of RR intervals are within the xx % highest or
    lowest values, i.e. if their z-score is above a threshold, e.g. ``1.96`` => 5% (2.5% highest, 2.5% lowest values);
    ``2.576`` => 1% (0.5% highest, 0.5% lowest values).


    Parameters
    ----------
    rpeaks : :class:`~pandas.DataFrame`
        dataframe with detected R peaks. Output from :meth:`biopsykit.signals.ecg.EcgProcessor.ecg_process()`
    bool_mask : :class:`numpy.array`
        boolean array with beats marked as outlier.
        Results of this outlier correction method will be combined with the array using a logical 'or'
    stat_thres : float
        threshold for z-score. Beats above that threshold will be marked as outlier


    Returns
    -------
    :class:`numpy.array`
        boolean array with beats marked as outlier. Logical 'or' combination of ``bool_mask`` and results from
        this algorithm

    """
    # statistical outlier: remove the x% highest and lowest successive differences of RR intervals
    # (1.96 std = 5% outlier, 2.576 std = 1% outlier)
    diff_rri = np.ediff1d(rpeaks["RR_Interval"], to_end=0)
    z_score = (diff_rri - np.nanmean(diff_rri)) / np.nanstd(diff_rri, ddof=1)

    return np.logical_or(bool_mask, np.abs(z_score) > stat_thres)


def _correct_outlier_artifact(
    rpeaks: pd.DataFrame,
    bool_mask: np.array,
    art_thres: float,  # noqa: ARG001
    **kwargs,  # noqa: ARG001
) -> np.array:
    """Apply outlier correction method 'artifact'.

    This function uses the artifact detection algorithm from `Berntson et al. (1990)`.
    Marks beats as outlier if they detected as such by this algorithm.


    Parameters
    ----------
    rpeaks : :class:`~pandas.DataFrame`
        dataframe with detected R peaks. Output from :meth:`biopsykit.signals.ecg.EcgProcessor.ecg_process()`
    bool_mask : :class:`numpy.array`
        boolean array with beats marked as outlier.
        Results of this outlier correction method will be combined with the array using a logical 'or'
    art_thres : float
        (not used but needed to ensure consistent method interface)


    Returns
    -------
    :class:`numpy.array`
        boolean array with beats marked as outlier. Logical 'or' combination of ``bool_mask`` and results from
        this algorithm


    References
    ----------
    Berntson, G. G., Quigley, K. S., Jang, J. F., & Boysen, S. T. (1990). An Approach to Artifact Identification:
    Application to Heart Period Data. *Psychophysiology*, 27(5), 586-598.
    https://doi.org/10.1111/j.1469-8986.1990.tb01982.x

    """
    # QD = Quartile Deviation = IQR / 2
    qd = iqr(rpeaks["RR_Interval"], nan_policy="omit") / 2.0
    # MAD = Minimal Artifact Difference
    mad = (rpeaks["RR_Interval"].median() - 2.9 * qd) / 3.0
    # MED = Maximum Expected Difference
    med = 3.32 * qd
    criterion = np.mean([mad, med])

    return np.logical_or(
        bool_mask,
        np.abs(rpeaks["RR_Interval"] - rpeaks["RR_Interval"].median()) > criterion,
    )


def _correct_outlier_physiological(
    rpeaks: pd.DataFrame, bool_mask: np.array, hr_thres: Tuple[float, float], **kwargs  # noqa: ARG001
) -> np.array:
    """Apply outlier correction method 'physiological'.

    This function marks beats as outlier if their heart rate is above or below a threshold that is very unlikely to be
    achieved physiologically.

    Parameters
    ----------
    rpeaks : :class:`~pandas.DataFrame`
        dataframe with detected R peaks. Output from :meth:`biopsykit.signals.ecg.EcgProcessor.ecg_process()`
    bool_mask : :class:`numpy.array`
        boolean array with beats marked as outlier.
        Results of this outlier correction method will be combined with the array using a logical 'or'
    hr_thres : tuple
        lower and upper heart rate thresholds. Beats below and above these values will be marked as outlier.


    Returns
    -------
    :class:`numpy.array`
        boolean array with beats marked as outlier. Logical 'or' combination of ``bool_mask`` and results from
        this algorithm

    """
    # physiological outlier: minimum/maximum heart rate threshold
    bool_mask = np.logical_or(
        bool_mask,
        (rpeaks["RR_Interval"] > (60 / hr_thres[0])) | (rpeaks["RR_Interval"] < (60 / hr_thres[1])),
    )
    return bool_mask


def _get_outlier_params(
    outlier_correction: Optional[Union[str, None, Sequence[str]]] = "all",
    outlier_params: Optional[Dict[str, Union[float, Sequence[float]]]] = None,
) -> Tuple[Sequence[str], Dict[str, Union[float, Sequence[float]]], Dict[str, Callable]]:

    if outlier_correction == "all":
        outlier_correction = list(_outlier_correction_methods.keys())
    elif isinstance(outlier_correction, str):
        outlier_correction = [outlier_correction]
    elif outlier_correction in ["None", None]:
        outlier_correction = []

    try:
        outlier_funcs: Dict[str, Callable] = {key: _outlier_correction_methods[key] for key in outlier_correction}
    except KeyError as e:
        raise ValueError(
            "`outlier_correction` may only contain values from {}, None or `all`, not `{}`.".format(
                list(_outlier_correction_methods.keys()), outlier_correction
            )
        ) from e

    if outlier_params is None:
        outlier_params = {key: _outlier_correction_params_default[key] for key in outlier_funcs}

    # get outlier params (values not passed as arguments will be filled with default arguments)
    outlier_params = {
        key: outlier_params[key] if key in outlier_params else _outlier_correction_params_default[key]
        for key in outlier_funcs
    }
    return outlier_correction, outlier_params, outlier_funcs


_hrv_methods = {
    "hrv_time": nk.hrv_time,
    "hrv_nonlinear": nk.hrv_nonlinear,
    "hrv_frequency": nk.hrv_frequency,
}

_edr_methods = {
    "peak_trough_mean": _edr_peak_trough_mean,
    "peak_trough_diff": _edr_peak_trough_diff,
    "peak_peak_interval": _edr_peak_peak_interval,
}

_outlier_correction_methods: Dict[str, Callable] = {
    "correlation": _correct_outlier_correlation,
    "quality": _correct_outlier_quality,
    "artifact": _correct_outlier_artifact,
    "physiological": _correct_outlier_physiological,
    "statistical_rr": _correct_outlier_statistical_rr,
    "statistical_rr_diff": _correct_outlier_statistical_rr_diff,
}

_outlier_correction_params_default: Dict[str, Union[float, Sequence[float]]] = {
    "correlation": 0.3,
    "quality": 0.4,
    "artifact": 0.0,
    "physiological": (45, 200),
    "statistical_rr": 2.576,
    "statistical_rr_diff": 1.96,
}


def _assert_rpeaks_input(
    ecg_processor: "EcgProcessor",
    key: str,
    rpeaks: RPeakDataFrame,
) -> None:
    """Assert valid input for ECG processing functions that require only R peaks.

    This function checks if either ``ecg_processor`` **and** ``key`` are supplied as arguments *or*
    ``rpeaks``.

    Parameters
    ----------
    ecg_processor : :class:`~biopsykit.signals.ecg.EcgProcessor`, optional
        ``EcgProcessor`` object. If this argument is supplied, the ``key`` argument needs to be supplied as well
    key : str, optional
        Dictionary key of the phase to process. Needed when ``ecg_processor`` is passed as argument
    rpeaks : :class:`~biopsykit.utils.datatype_helper.RPeakDataFrame`, optional
        Dataframe with detected R peaks. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process`

    Raises
    ------
    ValueError
        if input is invalid

    """
    if all(x is None for x in [ecg_processor, key]) and rpeaks is None:
        raise ValueError("Either 'ecg_processor' and 'key', or 'rpeaks' must be passed as arguments!")
    if ecg_processor is not None and key is None:
        raise ValueError("Both of 'ecg_processor' and 'key' must be passed as arguments!")
    if ecg_processor is None and rpeaks is None:
        raise ValueError("'rpeaks' must be passed as arguments when 'ecg_processor' is None!")


def _assert_ecg_input(ecg_processor: "EcgProcessor", key: str, ecg_signal: EcgResultDataFrame, rpeaks: RPeakDataFrame):
    """Assert valid input for ECG processing functions that require both only ECG signal and R peaks.

    This function checks if either ``ecg_processor`` **and** ``key`` are supplied as arguments *or*
    ``ecg_signal`` **and** `rpeaks`.

    Parameters
    ----------
    ecg_processor : :class:`~biopsykit.signals.ecg.EcgProcessor`, optional
        ``EcgProcessor`` object. If this argument is supplied, the ``key`` argument needs to be supplied as well
    key : str, optional
        Dictionary key of the phase to process. Needed when ``ecg_processor`` is passed as argument
    ecg_signal : :class:`~biopsykit.utils.datatype_helper.EcgResultDataFrame`, optional
        Dataframe with processed ECG signal. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`
    rpeaks : :class:`~biopsykit.utils.datatype_helper.RPeakDataFrame`, optional
        Dataframe with detected R peaks. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`

    Raises
    ------
    ValueError
        if input is invalid

    """
    if all(x is None for x in [ecg_processor, key]) and all(x is None for x in [ecg_signal, rpeaks]):
        raise ValueError("Either 'ecg_processor' and 'key', or 'rpeaks' and 'ecg_signal' must be passed as arguments!")
    if ecg_processor is not None and key is None:
        raise ValueError("Both of 'ecg_processor' and 'key' must be passed as arguments!")
    if ecg_signal is not None and rpeaks is None:
        raise ValueError("Both of 'ecg_signal' and 'rpeaks' must be passed as arguments!")


def _get_imputation_type(imputation_type: str, imputation_types: Sequence[str]) -> str:
    if imputation_type is None:
        imputation_type = "moving_average"
    elif imputation_type not in imputation_types:
        raise ValueError(f"'imputation_type' must be one of {imputation_types}, not {imputation_type}!")
    return imputation_type


def _check_dataframe_format(ecg_signal: EcgResultDataFrame, rpeaks: RPeakDataFrame):
    if ecg_signal is not None:
        is_ecg_result_dataframe(ecg_signal)
    is_r_peak_dataframe(rpeaks)



"""Utility functions for the ``biopsykit.protocols`` module."""
from typing import Dict, Optional, Sequence, Union

import numpy as np
import pandas as pd
from biopsykit.utils._datatype_validation_helper import _assert_len_list
from biopsykit.utils.datatype_helper import SalivaMeanSeDataFrame
from biopsykit.utils.exceptions import ValidationError
from pandas._libs.lib import is_timedelta_or_timedelta64_array


def _get_sample_times(
    saliva_data: Union[Dict[str, SalivaMeanSeDataFrame], SalivaMeanSeDataFrame],
    sample_times: Union[Sequence[int], Dict[str, Sequence[int]]],
    test_times: Sequence[int],
    sample_times_absolute: Optional[bool] = False,
) -> Union[Sequence[int], Dict[str, Sequence[int]]]:

    if isinstance(sample_times, dict):
        for key in sample_times:
            sample_times[key] = _get_sample_times(saliva_data[key], sample_times[key], test_times)
        return sample_times

    if isinstance(saliva_data, dict):
        sample_times = {}
        for key in saliva_data:
            sample_times[key] = _get_sample_times(saliva_data[key], sample_times, test_times)
        return sample_times

    if sample_times is None:
        sample_times = _get_sample_times_extract(saliva_data)

    _assert_len_list(test_times, 2)
    # ensure numpy
    sample_times = np.array(sample_times)

    if not sample_times_absolute:
        sample_times = _get_sample_times_absolute(sample_times, test_times)
    return list(sample_times)


def _get_sample_times_absolute(sample_times: np.ndarray, test_times: Sequence[int]):
    if is_timedelta_or_timedelta64_array(sample_times.flatten()):
        # convert into minutes
        sample_times_idx = sample_times.astype(float) / (1e9 * 60)
    else:
        sample_times_idx = sample_times
    index_post = np.where(sample_times_idx >= test_times[0])[0]
    sample_times[index_post] = sample_times[index_post] + (test_times[1] - test_times[0])
    return sample_times


def _get_sample_times_extract(saliva_data: pd.DataFrame):
    if "time" in saliva_data.reset_index().columns:
        return saliva_data.reset_index()["time"].unique()
    raise ValueError(
        "No sample times specified! Sample times must either be specified by passing them to the "
        "'sample_times' parameter or to by including them in 'saliva_data' ('time' column)."
    )


def _check_sample_times_match(data: SalivaMeanSeDataFrame, sample_times: Sequence[int]) -> None:
    if not any(len(sample_times) == s for s in [len(data.index.get_level_values("sample").unique()), len(data)]):
        raise ValidationError(
            "Number of saliva samples in data does not match number of samples in 'sample_times'. "
            "Expected {}, got {}.".format(len(data), len(sample_times))
        )



"""Module for processing Respiration data."""
from typing import Dict, Optional, Sequence, Union

import neurokit2 as nk
import numpy as np
import pandas as pd
import scipy.signal as ss
from biopsykit.signals._base import _BaseProcessor
from biopsykit.utils.array_handling import sanitize_input_1d

__all__ = ["RspProcessor"]


class RspProcessor(_BaseProcessor):
    """Class for processing Respiration data."""

    def __init__(
        self,
        data: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
        sampling_rate: Optional[float] = None,
        time_intervals: Optional[Union[pd.Series, Dict[str, Sequence[str]]]] = None,
        include_start: Optional[bool] = False,
    ):
        """Initialize an ``RspProcessor`` instance.

        You can either pass a data dictionary 'data_dict' containing Respiration data or dataframe containing
        Respiration data. For the latter, you can additionally supply time information via ``time_intervals`` parameter
        to automatically split the data into single phases.


        Parameters
        ----------
        data : :class:`~pandas.DataFrame` or dict
            dataframe (or dict of such) with Respiration data
        sampling_rate : float, optional
            sampling rate of recorded data
        time_intervals : dict or :class:`~pandas.Series`, optional
            time intervals indicating how ``data`` should be split.
            Can either be a :class:`~pandas.Series` with the `start` times of the single phases
            (the phase names are then derived from the index) or a dictionary with tuples indicating
            `start` and `end` times of phases (the phase names are then derived from the dict keys).
            Default: ``None`` (data is not split further)
        include_start : bool, optional
            ``True`` to include the data from the beginning of the recording to the first time interval as the
            first phase (then named ``Start``), ``False`` otherwise. Default: ``False``

        """
        super().__init__(
            data=data, sampling_rate=sampling_rate, time_intervals=time_intervals, include_start=include_start
        )

    @classmethod
    def rsp_compute_rate(cls, rsp_signal: pd.DataFrame, sampling_rate: Optional[float] = 256.0) -> float:
        """Compute respiration rate for given interval from respiration signal.

        Parameters
        ----------
        rsp_signal : :class:`~pandas.DataFrame`
            Raw respiration signal (1D). Can be a 'true' respiration signal (e.g. from bioimpedance or Radar)
            or an 'estimated' respiration signal (e.g. from ECG-derived respiration)
        sampling_rate : float, optional
            Sampling rate of recorded data


        Returns
        -------
        float
            Respiration rate during the given interval in bpm (breaths per minute)


        References
        ----------
        Schäfer, A., & Kratky, K. W. (2008). Estimation of Breathing Rate from Respiratory Sinus Arrhythmia:
        Comparison of Various Methods. *Annals of Biomedical Engineering*, 36(3), 476-485.
        https://doi.org/10.1007/s10439-007-9428-1


        Examples
        --------
        >>> from biopsykit.signals.ecg import EcgProcessor
        >>> # initialize EcgProcessor instance
        >>> ecg_processor = EcgProcessor(...)

        >>> # Extract respiration signal estimated from ECG using the 'peak_trough_diff' method
        >>> rsp_signal = ecg_processor.ecg_estimate_rsp(ecg_processor, key="Data", edr_type='peak_trough_diff')
        >>> # Compute respiration rate from respiration signal
        >>> rsp_rate = ecg_processor.rsp_compute_rate(rsp_signal)

        """
        # find peaks: minimal distance between peaks: 1 seconds
        rsp_signal = sanitize_input_1d(rsp_signal)
        edr_maxima = ss.find_peaks(rsp_signal, height=0, distance=sampling_rate)[0]
        edr_minima = ss.find_peaks(-1 * rsp_signal, height=0, distance=sampling_rate)[0]
        # threshold: 0.2 * Q3 (= 75th percentile)
        max_threshold = 0.2 * np.percentile(rsp_signal[edr_maxima], 75)
        # find all maxima that are above the threshold
        edr_maxima = edr_maxima[rsp_signal[edr_maxima] > max_threshold]

        # rearrange maxima into 2D array to that each row contains the start and end of one cycle
        rsp_cycles_start_end = np.vstack([edr_maxima[:-1], edr_maxima[1:]]).T

        # check for
        valid_resp_phases_mask = np.apply_along_axis(cls._check_contains_trough, 1, rsp_cycles_start_end, edr_minima)
        rsp_cycles_start_end = rsp_cycles_start_end[valid_resp_phases_mask]

        edr_signal_split: list = np.split(rsp_signal, rsp_cycles_start_end.flatten())[1::2]
        rsp_cycles_start_end = rsp_cycles_start_end.T
        rsp_peaks = rsp_cycles_start_end[0]
        rsp_troughs = np.array(list(map(np.argmin, edr_signal_split))) + rsp_cycles_start_end[0]

        rsp_rate_peaks = cls._rsp_rate(rsp_peaks, int(sampling_rate), len(rsp_signal))
        rsp_rate_troughs = cls._rsp_rate(rsp_troughs, int(sampling_rate), len(rsp_signal))
        return np.concatenate([rsp_rate_peaks, rsp_rate_troughs]).mean()

    @classmethod
    def _check_contains_trough(cls, start_end: np.ndarray, minima: np.ndarray) -> bool:
        """Check whether exactly one minima is in the interval given by the array ``start_end``: [start, end].

        Parameters
        ----------
        start_end : :class:`numpy.array`
            Array with start and end index
        minima : :class:`numpy.array`
            Array containing minima to be checked


        Returns
        -------
        bool
            ``True`` if exactly one minima is in the ``[start, end]`` interval, ``False`` otherwise

        """
        start, end = start_end
        return minima[(minima > start) & (minima < end)].shape[0] == 1

    @classmethod
    def _rsp_rate(cls, extrema: np.array, sampling_rate: int, desired_length: int) -> np.array:
        """Compute continuous respiration rate from extrema values.

        Parameters
        ----------
        extrema: :class:`numpy.array`
            List of respiration extrema (peaks or troughs)
        sampling_rate : float
            Sampling rate of recorded data
        desired_length : int
            Desired length of the output signal


        Returns
        -------
        :class:`numpy.array`
            Respiration rate array interpolated to desired length

        """
        rsp_rate_raw = (sampling_rate * 60) / np.ediff1d(extrema)
        # remove last sample
        x_old = extrema[:-1]
        x_new = np.linspace(x_old[0], x_old[-1], desired_length)
        return nk.signal_interpolate(x_old, rsp_rate_raw, x_new, method="linear")



"""Module providing helper functions to convert old BioPsyKit export files to new export formats."""
from typing import Optional

from biopsykit.io.ecg import load_hr_phase_dict, write_hr_phase_dict
from biopsykit.utils._types import path_t

__all__ = ["legacy_convert_hr_phase_dict"]

from biopsykit.utils.datatype_helper import HeartRatePhaseDict


def legacy_convert_hr_phase_dict(file_path: path_t, export: Optional[bool] = True) -> Optional[HeartRatePhaseDict]:
    """Legacy conversion for :obj:`~biopsykit.utils.datatype_helper.HeartRatePhaseDict`.

    Legacy conversion includes changing the column name from "ECG_Rate" into "Heart_Rate".


    Parameters
    ----------
    file_path : :class:`~pathlib.Path` or str
        path to file
    export : bool, optional
        ``True`` to directly export the converted
        :obj:`~biopsykit.utils.datatype_helper.HeartRatePhaseDict` under the same file name,
        ``False`` to return the converted :obj:`~biopsykit.utils.datatype_helper.HeartRatePhaseDict`


    Returns
    -------
    :obj:`~biopsykit.utils.datatype_helper.HeartRatePhaseDict`
        The converted :obj:`~biopsykit.utils.datatype_helper.HeartRatePhaseDict`, if ``export`` is ``False``

    """
    hr_phase_dict = load_hr_phase_dict(file_path, assert_format=False)
    for key, df in hr_phase_dict.items():
        df.columns = ["Heart_Rate"]
        hr_phase_dict[key] = df
    if export:
        write_hr_phase_dict(hr_phase_dict, file_path)
        return None
    return hr_phase_dict



"""A set of util functions to detect static regions in a IMU signal given certain constrains."""
from typing import Literal, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from biopsykit.utils._types import arr_t
from biopsykit.utils.array_handling import (
    _bool_fill,
    bool_array_to_start_end_array,
    sanitize_input_nd,
    sanitize_sliding_window_input,
    sliding_window_view,
)
from numpy.linalg import norm

# supported metric functions
_METRIC_FUNCTIONS = {
    "maximum": np.nanmax,
    "variance": np.nanvar,
    "mean": np.nanmean,
    "median": np.nanmedian,
}
METRIC_FUNCTION_NAMES = Literal["maximum", "variance", "mean", "median"]


def _find_static_samples(
    data: np.ndarray,
    window_length: int,
    inactive_signal_th: float,
    metric: METRIC_FUNCTION_NAMES = "mean",
    overlap: int = None,
) -> np.ndarray:
    """Search for static samples within given input signal, based on windowed L2-norm thresholding.

    .. warning::
        Due to edge cases at the end of the input data where window size and overlap might not fit your data, the last
        window might be discarded for analysis and will therefore always be considered as non-static!


    Parameters
    ----------
    data : array with shape (n, 3) or (n, 2)
        3D or 2D signal on which static moment detection should be performed (e.g. 3D-acc or 3D-gyr data)
    window_length : int
        Length of desired window in units of samples
    inactive_signal_th : float
       Threshold to decide whether a window should be considered as active or inactive. Window will be tested on
       <= threshold
    metric : str, optional
        Metric which will be calculated per window, one of the following strings:

        * 'mean' (default)
          Calculates mean value per window
        * 'maximum'
          Calculates maximum value per window
        * 'median'
          Calculates median value per window
        * 'variance'
          Calculates variance value per window

    overlap : int, optional
        Length of desired overlap in units of samples. If ``None`` (default) overlap will be window_length - 1


    Returns
    -------
    Boolean array with length n to indicate static (=``True``) or non-static (=``False``) for each sample


    Examples
    --------
    >>> _find_static_samples(data, window_length=128, overlap=64, inactive_signal_th = 5, metric = 'mean')

    See Also
    --------
    :func:`~biopsykit.utils.array_handling.sliding_window_view`
        Details on the used windowing function for this method.

    """
    # test for correct input data shape
    if np.shape(data)[-1] != 3 and np.shape(data)[-1] != 2:
        raise ValueError("Invalid signal dimensions, signal must be of shape (n,3) or (n, 2).")

    if metric not in _METRIC_FUNCTIONS:
        raise ValueError(f"Invalid metric passed! {metric} as metric is not supported.")

    # add default overlap value
    if overlap is None:
        overlap = window_length - 1

    # allocate output array
    inactive_signal_bool_array = np.zeros(len(data))

    # calculate norm of input signal (do this outside of loop to boost performance at cost of memory!)
    signal_norm = norm(data, axis=1)

    mfunc = _METRIC_FUNCTIONS[metric]

    # Create windowed view of norm
    windowed_norm = sliding_window_view(signal_norm, window_length, overlap, nan_padding=False)
    is_static = np.broadcast_to(mfunc(windowed_norm, axis=1) <= inactive_signal_th, windowed_norm.shape[::-1]).T

    # create the list of indices for sliding windows with overlap
    windowed_indices = sliding_window_view(np.arange(0, len(data)), window_length, overlap, nan_padding=False)

    # iterate over sliding windows
    inactive_signal_bool_array = _bool_fill(windowed_indices, is_static, inactive_signal_bool_array)

    return inactive_signal_bool_array.astype(bool)


def _find_static_sequences(
    data: np.ndarray,
    window_length: int,
    inactive_signal_th: float,
    metric: METRIC_FUNCTION_NAMES = "variance",
    overlap: int = None,
) -> np.ndarray:
    """Search for static sequences within given input signal, based on windowed L2-norm thresholding.

    .. warning::
        Due to edge cases at the end of the input data where window size and overlap might not fit your data, the last
        window might be discarded for analysis and will therefore always be considered as non-static!

    Parameters
    ----------
    data : array with shape (n, 3) or (n, 2)
        3D or 2D signal on which static moment detection should be performed (e.g. 3D-acc or 3D-gyr data)
    window_length : int
        Length of desired window in units of samples
    inactive_signal_th : float
       Threshold to decide whether a window should be considered as active or inactive. Window will be tested on
       <= threshold
    metric : str, optional
        Metric which will be calculated per window, one of the following strings:
            * 'variance' (default): Calculates variance value per window
            * 'mean': Calculates mean value per window
            * 'maximum': Calculates maximum value per window
            * 'median': Calculates median value per window

    overlap : int, optional
        Length of desired overlap in units of samples. If None (default) overlap will be window_length - 1

    Returns
    -------
    Array of [start, end] labels indication static regions within the input signal

    Examples
    --------
    >>> _find_static_sequences(data, window_length=128, overlap=64, inactive_signal_th = 5, metric = 'mean')

    See Also
    --------
    :func:`~biopsykit.signals.utils.array_handling.sliding_window`
        Details on the used windowing function for this method.

    """
    static_moment_bool_array = _find_static_samples(
        data=data, window_length=window_length, inactive_signal_th=inactive_signal_th, metric=metric, overlap=overlap
    )
    return bool_array_to_start_end_array(static_moment_bool_array)


def find_static_moments(
    data: arr_t,
    threshold: float,
    window_samples: Optional[int] = None,
    window_sec: Optional[int] = None,
    sampling_rate: Optional[Union[int, float]] = 0,
    overlap_samples: Optional[int] = None,
    overlap_percent: Optional[float] = None,
    metric: METRIC_FUNCTION_NAMES = "variance",
) -> pd.DataFrame:
    """Search for static moments within given input signal, based on windowed L2-norm thresholding.

    The window size of sliding windows can either be specified in *samples* (``window_samples``)
    or in *seconds* (``window_sec``, together with ``sampling_rate``).

    The overlap of windows can either be specified in *samples* (``overlap_samples``)
    or in *percent* (``overlap_percent``).


    Parameters
    ----------
    data : array with shape (n, 3) or (n, 2)
        3D or 2D signal on which static moment detection should be performed (e.g. 3D-acc or 3D-gyr data)
    window_samples : int, optional
        window size in samples or ``None`` if window size is specified in seconds + sampling rate. Default: ``None``
    window_sec : int, optional
        window size in seconds or ``None`` if window size is specified in samples. Default: ``None``
    sampling_rate : float, optional
        sampling rate of data in Hz. Only needed if window size is specified in seconds (``window_sec`` parameter).
        Default: ``None``
    overlap_samples : int, optional
        overlap of windows in samples or ``None`` if window overlap is specified in percent. Default: ``None``
    overlap_percent : float, optional
        overlap of windows in percent or ``None`` if window overlap is specified in samples. Default: ``None``
    threshold : float
       Threshold to decide whether a window should be considered as active or inactive. Window will be tested on
       <= threshold
    metric : str, optional
        Metric which will be calculated per window, one of the following strings:

        * 'variance' (default): Calculates variance value per window
        * 'mean': Calculates mean value per window
        * 'maximum': Calculates maximum value per window
        * 'median': Calculates median value per window


    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with ["start", "end"] columns indicating beginning and end of static regions within the input signal


    Examples
    --------
    >>> _find_static_sequences(data, window_length=128, overlap=64, inactive_signal_th = 5, metric = 'mean')


    See Also
    --------
    :func:`~biopsykit.utils.array_handling.sliding_window`
        Details on the used windowing function for this method.

    """
    # compute the data_norm of the variance in the windows
    window, overlap = sanitize_sliding_window_input(
        window_samples=window_samples,
        window_sec=window_sec,
        sampling_rate=sampling_rate,
        overlap_samples=overlap_samples,
        overlap_percent=overlap_percent,
    )
    if data.empty:
        start_end = np.zeros(shape=(0, 2))
    else:
        data = sanitize_input_nd(data)
        start_end = _find_static_sequences(
            data, window_length=window, overlap=overlap, inactive_signal_th=threshold, metric=metric
        )
        if len(start_end) == 0:
            return pd.DataFrame(columns=["start", "end"])
        # end indices are *inclusive*!
        start_end[:, 1] -= 1
    return pd.DataFrame(start_end, columns=["start", "end"])


def find_first_static_window_multi_sensor(
    signals: Sequence[np.ndarray],
    window_length: int,
    inactive_signal_th: float,
    metric: METRIC_FUNCTION_NAMES,
) -> Tuple[int, int]:
    """Find the first time window in the signal where all provided sensors are static.

    Parameters
    ----------
    signals : Sequence of n arrays with shape (k, m) or a 3D-array with shape (k, n, m)
        The signals of n senors with m axis and k samples.
    window_length
        Length of the required static signal in samples
    inactive_signal_th
        The threshold for static windows.
        If metric(norm(window, axis=-1)) <= `inactive_signal_th` for all sensors, it is considered static.
    metric
        The metric that should be calculated on the vectornorm over all axis for each sensor in each window

    Returns
    -------
    (start, end)
        Start and end index of the first static window.

    Examples
    --------
    >>> sensor_1_gyro = ...
    >>> sensor_2_gyro = ...
    >>> find_first_static_window_multi_sensor([sensor_1_gyro, sensor_2_gyro], window_length=128, inactive_signal_th=5)

    """
    if metric not in _METRIC_FUNCTIONS:
        raise ValueError(f"`metric` must be one of {list(_METRIC_FUNCTIONS.keys())}")

    if not isinstance(signals, np.ndarray):
        # all signals should have the same shape
        if not all(signals[0].shape == signal.shape for signal in signals):
            raise ValueError("All provided signals need to have the same shape.")
        if signals[0].ndim != 2:
            raise ValueError(
                "The array of each sensor must be 2D, where the first dimension is the time and the second dimension "
                "the sensor axis."
            )
        signals = np.hstack(signals)
    elif signals.ndim != 3:
        raise ValueError(
            "If a array is used as input, it must be 3D, where the first dimension is the time, "
            "the second indicates the sensor and the third the axis of the sensor."
        )

    n_signals = signals.shape[1]

    windows = sliding_window_view(
        signals.reshape((signals.shape[0], -1)),
        window_length=window_length,
        overlap=window_length - 1,
        nan_padding=False,
    )
    reshaped_windows = windows.reshape((*windows.shape[:-1], n_signals, -1))
    window_norm = norm(reshaped_windows, axis=-1)

    method = _METRIC_FUNCTIONS[metric]
    # This is pretty wasteful as we calculate the the function on all windows, even though we are only interested in
    # the first, where our threshold is valid.
    window_over_thres = method(window_norm, axis=1).max(axis=-1) <= inactive_signal_th

    valid_windows = np.nonzero(window_over_thres)[0]
    if len(valid_windows) == 0:
        raise ValueError("No static window was found")

    return valid_windows[0], valid_windows[0] + window_length



"""A couple of helper functions that ease the use of the typical biopsykit data formats."""
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd
from biopsykit.utils._datatype_validation_helper import (
    _assert_has_column_multiindex,
    _assert_has_column_prefix,
    _assert_has_columns,
    _assert_has_columns_any_level,
    _assert_has_index_levels,
    _assert_has_multiindex,
    _assert_is_dtype,
)
from biopsykit.utils.exceptions import ValidationError

__all__ = [
    "ECG_RESULT_DATAFRAME_COLUMNS",
    "HEART_RATE_DATAFRAME_COLUMNS",
    "R_PEAK_DATAFRAME_COLUMNS",
    "CodebookDataFrame",
    "MeanSeDataFrame",
    "BiomarkerRawDataFrame",
    "SalivaRawDataFrame",
    "SalivaFeatureDataFrame",
    "SalivaMeanSeDataFrame",
    "SleepEndpointDataFrame",
    "SleepEndpointDict",
    "EcgRawDataFrame",
    "EcgResultDataFrame",
    "RPeakDataFrame",
    "HeartRateDataFrame",
    "Acc1dDataFrame",
    "Acc3dDataFrame",
    "Gyr1dDataFrame",
    "Gyr3dDataFrame",
    "ImuDataFrame",
    "SleepWakeDataFrame",
    "SubjectConditionDataFrame",
    "SubjectConditionDict",
    "PhaseDict",
    "SubjectDataDict",
    "HeartRatePhaseDict",
    "HeartRateSubjectDataDict",
    "HeartRateStudyDataDict",
    "StudyDataDict",
    "MergedStudyDataDict",
    "is_subject_condition_dataframe",
    "is_subject_condition_dict",
    "is_codebook_dataframe",
    "is_mean_se_dataframe",
    "is_phase_dict",
    "is_hr_phase_dict",
    "is_subject_data_dict",
    "is_hr_subject_data_dict",
    "is_study_data_dict",
    "is_merged_study_data_dict",
    "is_saliva_raw_dataframe",
    "is_saliva_feature_dataframe",
    "is_saliva_mean_se_dataframe",
    "is_sleep_endpoint_dataframe",
    "is_sleep_endpoint_dict",
    "is_ecg_raw_dataframe",
    "is_ecg_result_dataframe",
    "is_r_peak_dataframe",
    "is_heart_rate_dataframe",
    "is_acc1d_dataframe",
    "is_acc3d_dataframe",
    "is_gyr1d_dataframe",
    "is_gyr3d_dataframe",
    "is_imu_dataframe",
    "is_sleep_wake_dataframe",
]

ECG_RESULT_DATAFRAME_COLUMNS = ["ECG_Raw", "ECG_Clean", "ECG_Quality", "ECG_R_Peaks", "R_Peak_Outlier"]
HEART_RATE_DATAFRAME_COLUMNS = ["Heart_Rate"]
R_PEAK_DATAFRAME_COLUMNS = ["R_Peak_Quality", "R_Peak_Idx", "RR_Interval", "R_Peak_Outlier"]

# these subclasses of pd.DataFrame are needed to be added to the type aliases because otherwise, autosphinx does not
# add the docstring to the documentation of the type aliases. Additionally, they can be used internally to highlight
# which alias types are expected at which position


class _SubjectConditionDataFrame(pd.DataFrame):
    pass


class _CodebookDataFrame(pd.DataFrame):
    pass


class _MeanSeDataFrame(pd.DataFrame):
    pass


class _BiomarkerRawDataFrame(pd.DataFrame):
    pass


class _SalivaRawDataFrame(pd.DataFrame):
    pass


class _SalivaFeatureDataFrame(pd.DataFrame):
    pass


class _SalivaMeanSeDataFrame(pd.DataFrame):
    pass


class _EcgRawDataFrame(pd.DataFrame):
    pass


class _EcgResultDataFrame(pd.DataFrame):
    pass


class _HeartRateDataFrame(pd.DataFrame):
    pass


class _RPeakDataFrame(pd.DataFrame):
    pass


class _Acc1dDataFrame(pd.DataFrame):
    pass


class _Acc3dDataFrame(pd.DataFrame):
    pass


class _Gyr1dDataFrame(pd.DataFrame):
    pass


class _Gyr3dDataFrame(pd.DataFrame):
    pass


class _ImuDataFrame(pd.DataFrame):
    pass


class _SleepWakeDataFrame(pd.DataFrame):
    pass


class _SleepEndpointDataFrame(pd.DataFrame):
    pass


SubjectConditionDataFrame = Union[_SubjectConditionDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing subject IDs and condition assignment in a standardized format.

A ``SubjectConditionDataFrame`` has an index with subject IDs named ``subject`` and a column with the condition
assignment named ``condition``.
"""

SubjectConditionDict = Dict[str, np.ndarray]
"""Dictionary containing subject IDs and condition assignment in a standardized format.

A ``SubjectConditionDict`` contains conditions as dictionary keys and a collection of subject IDs
(list, numpy array, pandas Index) as dictionary values.
"""

CodebookDataFrame = Union[_CodebookDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` representing a codebook which encodes numerical and categorical values
in a standardized format.

A ``CodebookDataFrame`` has an index level named ``variable``. The column names are the numerical values (0, 1, ...),
the dataframe entries then represent the mapping of numerical value to categorical value for the variable.
"""

MeanSeDataFrame = Union[_MeanSeDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing mean and standard error of time-series data in a standardized format.

The resulting dataframe must at least the two columns ``mean`` and ``se``. It can have additional index levels,
such as ``phase``, ``subphase`` or ``condition``.
"""

BiomarkerRawDataFrame = Union[_BiomarkerRawDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing raw biomarker data in a standardized format.

Data needs to be in long-format and **must** have a :class:`pandas.MultiIndex` with index level names:

* ``subject``: subject ID; can be number or string
* ``sample``: saliva sample ID; can be number or string

Additionally, the following index levels can be added to identify saliva values, such as:

* ``condition``: subject condition during the study (e.g., "Control" vs. "Condition")
* ``day``: day ID, if saliva samples were collected over multiple days
* ``night``: night ID, if saliva samples were collected over multiple night
* ...

"""

SalivaRawDataFrame = Union[_BiomarkerRawDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing raw saliva data in a standardized format.

Data needs to be in long-format and **must** have a :class:`pandas.MultiIndex` with index level names:

* ``subject``: subject ID; can be number or string
* ``sample``: saliva sample ID; can be number or string

Additionally, the following index levels can be added to identify saliva values, such as:

* ``condition``: subject condition during the study (e.g., "Control" vs. "Condition")
* ``day``: day ID, if saliva samples were collected over multiple days
* ``night``: night ID, if saliva samples were collected over multiple night
* ...

"""

SalivaFeatureDataFrame = Union[_SalivaFeatureDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing feature computed from saliva data in a standardized format.

The resulting dataframe must at least have a ``subject`` index level and all column names need to begin with
the saliva marker type (e.g. "cortisol"), followed by the feature name, separated by underscore '_'
Additionally, the name of the column index needs to be `saliva_feature`.
"""

SalivaMeanSeDataFrame = Union[_SalivaMeanSeDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing mean and standard error of saliva samples in a standardized format.

The resulting dataframe must at least have a ``sample`` index level and the two columns ``mean`` and ``se``.
It can have additional index levels, such as ``condition`` or ``time``.
"""

SleepEndpointDict = Dict[str, Any]
"""Dictionary containing sleep endpoints in a standardized format.

The dict entries represent the sleep endpoints and should follow a standardized naming convention,
regardless of the origin (IMU sensor, sleep mattress, psg, etc.).

Required are the entries:

* ``sleep_onset``: Sleep Onset, i.e., time of falling asleep, in absolute time
* ``wake_onset``: Wake Onset, i.e., time of awakening, in absolute time
* ``total_sleep_duration``: Total sleep duration, i.e., time between Sleep Onset and Wake Onset, in minutes

The following entries are common, but not required:

* ``total_duration``: Total recording time, in minutes
* ``net_sleep_duration``: Net duration spent sleeping, in minutes
* ``bed_interval_start``: Bed Interval Start, i.e, time when participant went to bed, in absolute time
* ``bed_interval_end``: Bed Interval End, i.e, time when participant left bed, in absolute time
* ``sleep_efficiency``: Sleep Efficiency, defined as the ratio between net sleep duration and sleep duration
  in percent
* ``sleep_onset_latency``: Sleep Onset Latency, i.e., time in bed needed to fall asleep, in minutes
* ``getup_latency``: Get Up Latency, i.e., time in bed after awakening until getting up, in minutes
* ``wake_after_sleep_onset``: Wake After Sleep Onset (WASO), i.e., total time awake after falling asleep, in minutes
* ``sleep_bouts``: List with start and end times of sleep bouts
* ``wake_bouts``: List with start and end times of wake bouts
* ``number_wake_bouts``: Total number of wake bouts

The following entries are, for instance, further possible:

* ``total_time_light_sleep``: Total time of light sleep, in minutes
* ``total_time_deep_sleep``: Total time of deep sleep, in minutes
* ``total_time_rem_sleep``: Total time of REM sleep, in minutes
* ``total_time_awake``: Total time of being awake, in minutes
* ``count_snoring_episodes``: Total number of snoring episodes
* ``total_time_snoring``: Total time of snoring, in minutes
* ``heart_rate_avg``: Average heart rate during recording, in bpm
* ``heart_rate_min``: Minimum heart rate during recording, in bpm
* ``heart_rate_max``: Maximum heart rate during recording, in bpm

"""

SleepEndpointDataFrame = Union[_SleepEndpointDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing sleep endpoints in a standardized format.

The resulting dataframe must at least have a ``date`` index level,
and, optionally, further index levels like ``night``.

The columns defining the sleep endpoints should follow a standardized naming convention, regardless of the origin
(IMU sensor, sleep mattress, psg, etc.).

Required are the columns:

* ``sleep_onset``: Sleep Onset, i.e., time of falling asleep, in absolute time
* ``wake_onset``: Wake Onset, i.e., time of awakening, in absolute time
* ``total_sleep_duration``: Total sleep duration, i.e., time between Sleep Onset and Wake Onset, in minutes

The following columns are common, but not required:

* ``total_duration``: Total recording time, in minutes
* ``net_sleep_duration``: Net duration spent sleeping, in minutes
* ``bed_interval_start``: Bed Interval Start, i.e, time when participant went to bed, in absolute time
* ``bed_interval_end``: Bed Interval End, i.e, time when participant left bed, in absolute time
* ``sleep_efficiency``: Sleep Efficiency, defined as the ratio between net sleep duration and sleep duration
  in percent
* ``sleep_onset_latency``: Sleep Onset Latency, i.e., time in bed needed to fall asleep, in minutes
* ``getup_latency``: Get Up Latency, i.e., time in bed after awakening until getting up, in minutes
* ``wake_after_sleep_onset``: Wake After Sleep Onset (WASO), i.e., total time awake after falling asleep, in minutes
* ``number_wake_bouts``: Total number of wake bouts

The following columns are further possible:

* ``total_time_light_sleep``: Total time of light sleep, in minutes
* ``total_time_deep_sleep``: Total time of deep sleep, in minutes
* ``total_time_rem_sleep``: Total time of REM sleep, in minutes
* ``total_time_awake``: Total time of being awake, in minutes
* ``count_snoring_episodes``: Total number of snoring episodes
* ``total_time_snoring``: Total time of snoring, in minutes
* ``heart_rate_avg``: Average heart rate during recording, in bpm
* ``heart_rate_min``: Minimum heart rate during recording, in bpm
* ``heart_rate_max``: Maximum heart rate during recording, in bpm

"""


EcgRawDataFrame = Union[_EcgRawDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing raw ECG data of `one` subject.

The dataframe is expected to have the following columns:

* ``ecg``: Raw ECG signal

"""

EcgResultDataFrame = Union[_EcgResultDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing processed ECG data of `one` subject.

The dataframe is expected to have the following columns:

* ``ECG_Raw``: Raw ECG signal
* ``ECG_Clean``: Cleaned (filtered) ECG signal
* ``ECG_Quality``: ECG signal quality indicator in the range of [0, 1]
* ``ECG_R_Peaks``: 1.0 where R peak was detected in the ECG signal, 0.0 else
* ``R_Peak_Outlier``: 1.0 when a detected R peak was classified as outlier, 0.0 else
* ``Heart_Rate``: Computed Heart rate interpolated to signal length

"""

HeartRateDataFrame = Union[_HeartRateDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing heart rate time series data of `one` subject.

The dataframe is expected to have the following columns:

* ``Heart_Rate``: Heart rate data. Can either be instantaneous heart rate or resampled heart rate

"""

RPeakDataFrame = Union[_RPeakDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing R-peak locations of `one` subject extracted from ECG data.

The dataframe is expected to have the following columns:

* ``R_Peak_Quality``: Signal quality indicator (of the raw ECG signal) in the range of [0, 1]
* ``R_Peak_Idx``: Array index of detected R peak in the raw ECG signal
* ``RR_Interval``: Interval between the current and the successive R peak in seconds
* ``R_Peak_Outlier``: 1.0 when a detected R peak was classified as outlier, 0.0 else

"""

Acc1dDataFrame = Union[_Acc1dDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing 1-d acceleration data.

The dataframe is expected to have one of the following column sets:

* ["acc"]: one level column index
* ["acc_norm"]: one level column index

"""

Acc3dDataFrame = Union[_Acc3dDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing 3-d acceleration data.

The dataframe is expected to have one of the following column sets:

* ["acc_x", "acc_y", "acc_z"]: one level column index
* [("acc", "x"), ("acc", "y"), ("acc", "z")]: two-level column index, first level specifying the channel
  (acceleration), second level specifying the axes

"""

Gyr1dDataFrame = Union[_Gyr1dDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing 1-d gyroscope data.

The dataframe is expected to have one of the following column sets:

* ["gyr"]: one level column index
* ["gyr_norm"]: one level column index

"""

Gyr3dDataFrame = Union[_Gyr3dDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing 3-d gyroscope data.

The dataframe is expected to have one of the following column sets:

* ["gyr_x", "gyr_y", "gyr_z"]: one level column index
* [("gyr", "x"), ("gyr", "y"), ("gyr", "z")]: two-level column index, first level specifying the channel
  (gyroscope), second level specifying the axes

"""

ImuDataFrame = Union[_ImuDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing 6-d inertial measurement (IMU) (acceleration and gyroscope) data.

Hence, an ``ImuDataFrame`` must both be a ``AccDataFrame`` **and** a ``GyrDataFrame``.

The dataframe is expected to have one of the following column sets:

* ["acc_x", "acc_y", "acc_z", "gyr_x", "gyr_y", "gyr_z"]: one level column index
* [("acc", "x"), ("acc", "y"), ("acc", "z"), ("gyr", "x"), ("gyr", "y"), ("gyr", "z")]:
  two-level column index, first level specifying the channel (acceleration and gyroscope),
  second level specifying the axes

"""

SleepWakeDataFrame = Union[_SleepWakeDataFrame, pd.DataFrame]
""":class:`~pandas.DataFrame` containing sleep/wake predictions.

The dataframe is expected to have at least the following column(s):

* ["sleep_wake"]: sleep/wake predictions where 1 indicates sleep and 0 indicates wake

"""

PhaseDict = Dict[str, pd.DataFrame]
"""Dictionary containing general time-series data of **one single subject** split into **different phases**.

A ``PhaseDict`` is a dictionary with the following format:

{ "phase_1" : dataframe, "phase_2" : dataframe, ... }

Each ``dataframe`` is a :class:`~pandas.DataFrame` with the following format:

* Index: :class:`pandas.DatetimeIndex` with timestamps, name of index level: ``time``

"""

HeartRatePhaseDict = Dict[str, HeartRateDataFrame]
"""Dictionary containing time-series heart rate data of **one single subject** split into **different phases**.

A ``HeartRatePhaseDict`` is a dictionary with the following format:

{ "phase_1" : hr_dataframe, "phase_2" : hr_dataframe, ... }

Each ``hr_dataframe`` is a :class:`~pandas.DataFrame` with the following format:

* ``time`` Index: :class:`pandas.DatetimeIndex` with heart rate sample timestamps
* ``Heart_Rate`` Column: heart rate values

"""

SubjectDataDict = Dict[str, PhaseDict]
"""Dictionary representing time-series data from **multiple subjects** collected during a psychological protocol.

A ``SubjectDataDict`` is a nested dictionary with time-series data from multiple subjects, each containing data
from different phases. It is expected to have the level order `subject`, `phase`:

| {
|     "subject1" : { "phase_1" : dataframe, "phase_2" : dataframe, ... },
|     "subject2" : { "phase_1" : dataframe, "phase_2" : dataframe, ... },
|     ...
| }

This dictionary can, for instance, be rearranged to a :obj:`biopsykit.utils.datatype_helper.StudyDataDict`,
where the level order is reversed: `phase`, `subject`.
"""

HeartRateSubjectDataDict = Union[Dict[str, HeartRatePhaseDict], Dict[str, Dict[str, HeartRatePhaseDict]]]
"""Dictionary with time-series heart rate data from **multiple subjects** collected during a psychological protocol.

A ``HeartRateSubjectDataDict`` is a nested dictionary with time-series heart rate data from multiple subjects,
each containing data from different phases. It is expected to have the level order `subject`, `phase`:

| {
|     "subject1" : { "phase_1" : hr_dataframe, "phase_2" : hr_dataframe, ... },
|     "subject2" : { "phase_1" : hr_dataframe, "phase_2" : hr_dataframe, ... },
|     ...
| }

Each ``hr_dataframe`` is a :class:`~pandas.DataFrame` with the following format:

* ``time`` Index: :class:`pandas.DatetimeIndex` with heart rate sample timestamps
* ``Heart_Rate`` Column: heart rate values

This dictionary can, for instance, be rearranged to a :obj:`~biopsykit.utils.datatype_helper.HeartRateStudyDataDict`,
where the level order is reversed: `phase`, `subject`.

"""

StudyDataDict = Dict[str, Dict[str, pd.DataFrame]]
"""Dictionary with data from **multiple phases** collected during a psychological protocol.

A ``StudyDataDict`` is a nested dictionary with time-series data from multiple phases, each phase containing data
from different subjects. It is expected to have the level order `phase`, `subject`:

| {
|     "phase_1" : { "subject1" : dataframe, "subject2" : dataframe, ... },
|     "phase_2" : { "subject1" : dataframe, "subject2" : dataframe, ... },
|     ...
| }

This dict results from rearranging a :obj:`biopsykit.utils.datatype_helper.SubjectDataDict` by calling
:func:`~biopsykit.utils.data_processing.rearrange_subject_data_dict`.
"""


HeartRateStudyDataDict = Dict[str, Dict[str, HeartRateDataFrame]]
"""Dictionary with heart rate data from **multiple phases** collected during a psychological protocol.

A ``HeartRateStudyDataDict`` is a nested dictionary with time-series heart rate data from multiple phases,
each phase containing data from different subjects. It is expected to have the level order `phase`, `subject`:

| {
|     "phase_1" : { "subject1" : hr_dataframe, "subject2" : hr_dataframe, ... },
|     "phase_2" : { "subject1" : hr_dataframe, "subject2" : hr_dataframe, ... },
|     ...
| }

Each ``hr_dataframe`` is a :class:`~pandas.DataFrame` with the following format:

* ``time`` Index: :class:`pandas.DatetimeIndex` with heart rate sample timestamps
* ``Heart_Rate`` Column: heart rate values

This dict results from rearranging a :obj:`~biopsykit.utils.datatype_helper.HeartRateSubjectDataDict` by calling
:func:`~biopsykit.utils.data_processing.rearrange_subject_data_dict`.
"""

MergedStudyDataDict = Dict[str, pd.DataFrame]
"""Dictionary with merged time-series data of **multiple subjects**, split into **different phases**.

A ``MergedStudyDataDict`` is a dictionary with the following format:

| {
|     "phase_1" : merged_dataframe,
|     "phase_2" : merged_dataframe,
|     ...
| }

This dict results from merging the inner dictionary into one dataframe by calling
:func:`~biopsykit.utils.data_processing.merge_study_data_dict`.

.. note::
    Merging the inner dictionaries requires that the dataframes of all subjects have same length within each phase.

Each ``merged_dataframe`` is a :class:`~pandas.DataFrame` with the following format:

* Index: time. Name of index level: ``time``
* Columns: time series data per subject, each subject has its own column.
  Name of the column index level: ``subject``
"""


def is_subject_condition_dataframe(
    data: SubjectConditionDataFrame, raise_exception: Optional[bool] = True
) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``SubjectConditionDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SubjectConditionDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SubjectConditionDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_multiindex(data, expected=False)
        _assert_has_index_levels(data, index_levels=["subject"], match_atleast=False, match_order=True)
        _assert_has_columns(data, [["condition"]])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a SubjectConditionDataFrame. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_subject_condition_dict(data: SubjectConditionDict, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDict`.

    Parameters
    ----------
    data : dict
        dict to check if it is a ``SubjectConditionDict``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SubjectConditionDict``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SubjectConditionDict``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDict`
        dictionary format

    """
    try:
        _assert_is_dtype(data, dict)
        for val in data.values():
            _assert_is_dtype(val, (np.ndarray, list, pd.Index))
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a SubjectConditionDict. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_codebook_dataframe(data: CodebookDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.CodebookDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``CodebookDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value


    Returns
    -------
    ``True`` if ``data`` is a ``CodebookDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)


    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``CodebookDataFrame``


    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.CodebookDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_index_levels(data, index_levels="variable", match_atleast=True, match_order=False)
        if not np.issubdtype(data.columns.dtype, np.integer):
            raise ValidationError(
                "The dtypes of columns in a CodebookDataFrame are expected to be of type int, but it is {}.".format(
                    data.columns.dtype
                )
            )
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a CodebookDataFrame. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_mean_se_dataframe(data: MeanSeDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.MeanSeDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``MeanSeDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value


    Returns
    -------
    ``True`` if ``data`` is a ``MeanSeDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)


    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``MeanSeDataFrame``


    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.MeanSeDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        if data.columns.nlevels == 1:
            _assert_has_columns(data, [["mean", "se"]])
        else:
            _assert_has_columns_any_level(data, [["mean", "se"]])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a MeanSeDataFrame. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_hr_phase_dict(data: HeartRatePhaseDict, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether a dict is a :obj:`~biopsykit.utils.datatype_helper.HeartRatePhaseDict`.

    Parameters
    ----------
    data : dict
        dict to check
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``HeartRatePhaseDict``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``HeartRatePhaseDict``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.HeartRatePhaseDict`
        dictionary format

    """
    try:
        _assert_is_dtype(data, dict)
        for df in data.values():
            is_heart_rate_dataframe(df)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a HeartRatePhaseDict. "
                "The validation failed with the following error:\n\n{}\n"
                "HeartRatePhaseDicts in an old format can be converted into the new format using "
                "`biopsykit.utils.legacy_helper.legacy_convert_hr_phase_dict()`".format(str(e))
            ) from e
        return False
    return True


def is_phase_dict(data: PhaseDict, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether a dict is a :obj:`~biopsykit.utils.datatype_helper.PhaseDict`.

    Parameters
    ----------
    data : dict
        dict to check
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``PhaseDict``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``PhaseDict``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.PhaseDict`
        dictionary format

    """
    try:
        _assert_is_dtype(data, dict)
        for df in data.values():
            _assert_is_dtype(df, pd.DataFrame)
            _assert_has_multiindex(df, expected=False)
            _assert_has_column_multiindex(df, expected=False)
            _assert_has_index_levels(df, ["time"])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a PhaseDict. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_hr_subject_data_dict(data: HeartRateSubjectDataDict, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether a dict is a :obj:`~biopsykit.utils.datatype_helper.HeartRateSubjectDataDict`.

    Parameters
    ----------
    data : dict
        dict to check
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``HeartRateSubjectDataDict``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``HeartRateSubjectDataDict``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.HeartRateSubjectDataDict`
        dictionary format

    """
    try:
        _assert_is_dtype(data, dict)
        for data_dict in data.values():
            is_hr_phase_dict(data_dict)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a HeartRateSubjectDataDict. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_study_data_dict(data: StudyDataDict, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether a dict is a :obj:`~biopsykit.utils.datatype_helper.StudyDataDict`.

    Parameters
    ----------
    data : dict
        dict to check
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``StudyDataDict``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``StudyDataDict``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.StudyDataDict`
        dictionary format

    """
    try:
        _assert_is_dtype(data, dict)
        for data_dict in data.values():
            _assert_is_dtype(data_dict, dict)
            for df in data_dict.values():
                _assert_is_dtype(df, pd.DataFrame)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a StudyDataDict. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_subject_data_dict(data: SubjectDataDict, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether a dict is a :obj:`~biopsykit.utils.datatype_helper.SubjectDataDict`.

    Parameters
    ----------
    data : dict
        dict to check
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SubjectDataDict```
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SubjectDataDict``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.SubjectDataDict``
        dictionary format

    """
    try:
        _assert_is_dtype(data, dict)
        for data_dict in data.values():
            _assert_is_dtype(data_dict, dict)
            for df in data_dict.values():
                _assert_is_dtype(df, pd.DataFrame)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a SubjectDataDict. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_merged_study_data_dict(data: MergedStudyDataDict, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether a dict is a :obj:`~biopsykit.utils.datatype_helper.MergedStudyDataDict`.

    Parameters
    ----------
    data : dict
        dict to check
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``MergedStudyDataDict``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``MergedStudyDataDict``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.MergedStudyDataDict`
        dictionary format

    """
    try:
        _assert_is_dtype(data, dict)
        for df in data.values():
            _assert_is_dtype(df, pd.DataFrame)
            _assert_has_multiindex(df, expected=False)
            _assert_has_index_levels(df, ["time"])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a MergedStudyDataDict. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_biomarker_raw_dataframe(
    data: BiomarkerRawDataFrame, biomarker_type: Union[str, List[str]], raise_exception: Optional[bool] = True
) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.SalivaRawDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``SalivaRawDataFrame``
    saliva_type : str or list of str
        type of saliva data (or list of saliva types) in the dataframe, e.g., "cortisol" or "amylase"
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SalivaRawDataFrame```
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SalivaRawDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.SalivaRawDataFrame`
        dataframe format

    """
    try:
        if biomarker_type is None:
            raise ValidationError("`saliva_type` is None!")
        if isinstance(biomarker_type, str):
            biomarker_type = [biomarker_type]
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_multiindex(data, nlevels=2, nlevels_atleast=True)
        _assert_has_index_levels(data, index_levels=["subject", "sample"], match_atleast=True, match_order=False)
        _assert_has_columns(data, [biomarker_type, [*biomarker_type, "time"]])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a BiomarkerRawDataFrame. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_saliva_raw_dataframe(
    data: SalivaRawDataFrame, saliva_type: Union[str, List[str]], raise_exception: Optional[bool] = True
) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.SalivaRawDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``SalivaRawDataFrame``
    saliva_type : str or list of str
        type of saliva data (or list of saliva types) in the dataframe, e.g., "cortisol" or "amylase"
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SalivaRawDataFrame```
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SalivaRawDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.SalivaRawDataFrame`
        dataframe format

    """
    return is_biomarker_raw_dataframe(data, saliva_type, raise_exception)


def is_saliva_feature_dataframe(
    data: SalivaFeatureDataFrame, saliva_type: str, raise_exception: Optional[bool] = True
) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.SalivaFeatureDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``SalivaFeatureDataFrame``
    saliva_type : str or list of str
        type of saliva data in the dataframe, e.g., "cortisol" or "amylase"
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SalivaFeatureDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SalivaFeatureDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.SalivaFeatureDataFrame`
        dataframe format

    """
    try:
        if saliva_type is None:
            raise ValidationError("`saliva_type` is None!")
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_index_levels(data, index_levels="subject", match_atleast=True, match_order=False)
        # _assert_has_column_levels(data, column_levels="saliva_feature", match_atleast=True, match_order=False)
        _assert_has_column_prefix(data.columns, prefix=saliva_type)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a SalivaFeatureDataFrame. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_saliva_mean_se_dataframe(data: SalivaFeatureDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.SalivaMeanSeDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``SalivaMeanSeDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SalivaMeanSeDataFrame```
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SalivaMeanSeDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.SalivaMeanSeDataFrame``
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_index_levels(data, index_levels="sample", match_atleast=True, match_order=False)
        _assert_has_columns(data, [["mean", "se"]])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a SalivaMeanSeDataFrame. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_sleep_endpoint_dataframe(data: SleepEndpointDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.SleepEndpointDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``SleepEndpointDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SleepEndpointDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SleepEndpointDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.SleepEndpointDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_is_dtype(data.index, pd.DatetimeIndex)
        _assert_has_index_levels(data, index_levels="date", match_atleast=True, match_order=False)
        _assert_has_columns(data, columns_sets=[["sleep_onset", "wake_onset", "total_sleep_duration"]])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a SleepEndpointDataFrame. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_sleep_endpoint_dict(data: SleepEndpointDict, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dictionary is a :obj:`~biopsykit.utils.datatype_helper.SleepEndpointDict`.

    Parameters
    ----------
    data : dict
        data to check if it is a ``SleepEndpointDict``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SleepEndpointDict``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SleepEndpointDict``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.SleepEndpointDict`
        dictionary format

    """
    try:
        _assert_is_dtype(data, dict)
        expected_keys = ["date", "sleep_onset", "wake_onset", "total_sleep_duration"]
        if any(col not in list(data.keys()) for col in expected_keys):
            raise ValidationError(f"Not all of {expected_keys} are in the dictionary!")
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a SleepEndpointDict. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_ecg_raw_dataframe(data: EcgRawDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.EcgRawDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``EcgRawDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``EcgRawDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``EcgRawDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.EcgRawDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(data, columns_sets=[["ecg"]])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a EcgRawDataFrame. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_ecg_result_dataframe(data: EcgRawDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.EcgResultDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``EcgResultDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``EcgResultDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``EcgResultDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.EcgResultDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(
            data,
            columns_sets=[
                ECG_RESULT_DATAFRAME_COLUMNS,
                ECG_RESULT_DATAFRAME_COLUMNS + HEART_RATE_DATAFRAME_COLUMNS,
            ],
        )
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a EcgResultDataFrame. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_heart_rate_dataframe(data: HeartRateDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.HeartRateDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``HeartRateDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``HeartRateDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``HeartRateDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.HeartRateDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(data, columns_sets=[HEART_RATE_DATAFRAME_COLUMNS])
        _assert_has_multiindex(data, expected=False)
        _assert_has_column_multiindex(data, expected=False)
        _assert_has_index_levels(data, ["time"])
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a HeartRateDataFrame. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_r_peak_dataframe(data: EcgRawDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.RPeakDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``RPeakDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``RPeakDataFrame```
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``RPeakDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.RPeakDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(
            data,
            columns_sets=[
                ["R_Peak_Idx", "RR_Interval"],
                ["R_Peak_Quality", "R_Peak_Idx", "RR_Interval"],
                R_PEAK_DATAFRAME_COLUMNS,
            ],
        )
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a RPeakDataFrame. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_acc1d_dataframe(data: Acc3dDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.Acc1dDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``Acc1dDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``Acc1dDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``Acc1dDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.Acc1dDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(
            data,
            columns_sets=[["acc"], ["acc_norm"]],
        )
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a Acc1dDataFrame. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_acc3d_dataframe(data: Acc3dDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.Acc3dDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``Acc3dDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``Acc3dDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``Acc3dDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.Acc3dDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(
            data,
            columns_sets=[
                ["acc_x", "acc_y", "acc_z"],
                [("acc", "x"), ("acc", "y"), ("acc", "z")],
            ],
        )
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a Acc3dDataFrame. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_gyr1d_dataframe(data: Gyr3dDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.Gyr1dDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``Gyr1dDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``Gyr1dDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``Gyr1dDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.Gyr1dDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(
            data,
            columns_sets=[["gyr"], ["gyr_norm"]],
        )
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a Gyr1dDataFrame. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_gyr3d_dataframe(data: Gyr3dDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.Gyr3dDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``Gyr3dDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``Gyr3dDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``Gyr3dDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.Gyr3dDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(
            data,
            columns_sets=[
                ["gyr_x", "gyr_y", "gyr_z"],
                [("gyr", "x"), ("gyr", "y"), ("gyr", "z")],
            ],
        )
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a Gyr3dDataFrame. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_imu_dataframe(data: Gyr3dDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.ImuDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``ImuDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``ImuDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``ImuDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.ImuDataFrame`
        dataframe format

    """
    try:
        is_acc3d_dataframe(data, raise_exception=True)
        is_gyr3d_dataframe(data, raise_exception=True)
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a ImuDataFrame. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True


def is_sleep_wake_dataframe(data: SleepWakeDataFrame, raise_exception: Optional[bool] = True) -> Optional[bool]:
    """Check whether dataframe is a :obj:`~biopsykit.utils.datatype_helper.SleepWakeDataFrame`.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to check if it is a ``SleepWakeDataFrame``
    raise_exception : bool, optional
        whether to raise an exception or return a bool value

    Returns
    -------
    ``True`` if ``data`` is a ``SleepWakeDataFrame``
    ``False`` otherwise (if ``raise_exception`` is ``False``)

    Raises
    ------
    ValidationError
        if ``raise_exception`` is ``True`` and ``data`` is not a ``SleepWakeDataFrame``

    See Also
    --------
    :obj:`~biopsykit.utils.datatype_helper.SleepWakeDataFrame`
        dataframe format

    """
    try:
        _assert_is_dtype(data, pd.DataFrame)
        _assert_has_columns(data, [["sleep_wake"]])
        if not all(data["sleep_wake"].between(0, 1, inclusive=True)):
            raise ValidationError(
                "Invalid values for sleep/wake prediction! Sleep/wake scores are expected to be "
                "in the interval [0, 1]."
            )
    except ValidationError as e:
        if raise_exception is True:
            raise ValidationError(
                "The passed object does not seem to be a SleepWakeDataFrame. "
                "The validation failed with the following error:\n\n{}".format(str(e))
            ) from e
        return False
    return True
